import opengate as gate
from opengate.contrib.spect.spect_helpers import *
from opengate.actors.biasingactors import distance_dependent_angle_tolerance
from opengate.exception import fatal
from opengate import g4_units
import opengate.contrib.spect.ge_discovery_nm670 as nm670
import opengate.contrib.spect.siemens_intevo as intevo
from opengate.image import read_image_info
from opengate.utility import get_basename_and_extension
from opengate.sources.utility import set_source_energy_spectrum
import numpy as np
import matplotlib.pyplot as plt
import os
import json
from pathlib import Path


class ConfigBase:
    """
    Base class that provides generic to_dict and from_dict methods
    for configuration objects and equality operator __eq__.
    """

    def __eq__(self, other):
        """
        Generic comparison for ConfigBase and derived classes.
        Recursively compares attributes, handling special cases like functions and paths.
        """
        if not isinstance(other, self.__class__):
            return False

        # Compare all attributes
        for key, value1 in vars(self).items():
            if key == "spect_config":
                continue
            if key not in vars(other):
                print(f'Attribute "{key}" not in other')
                return False
            value2 = getattr(other, key)
            if not self.compare_values(value1, value2):
                print(f"Values for attribute '{key}' differ :" f" {value1} != {value2}")
                return False

        # Ensure consistent attributes in both objects
        if set(vars(self).keys()) != set(vars(other).keys()):
            return False

        return True

    @staticmethod
    def compare_values(value1, value2):
        """
        Compare two values, handling special cases like callables, paths, and ConfigBase objects.
        """
        # Handle function comparison by name and module
        if callable(value1) and callable(value2):
            return (
                value1.__name__ == value2.__name__
                and value1.__module__ == value2.__module__
            )
        # Handle Path objects comparison
        if isinstance(value1, Path) and isinstance(value2, Path):
            return str(value1) == str(value2)
        # Handle nested ConfigBase objects
        if isinstance(value1, ConfigBase) and isinstance(value2, ConfigBase):
            return value1 == value2
        # Default comparison
        return value1 == value2

    def serialize_value(self, value):
        """Helper method to serialise different types of values."""
        if callable(value):
            return {
                "type": "function",
                "module": value.__module__,
                "function_name": value.__name__,
            }
        if isinstance(value, Path):
            return str(value)
        if isinstance(value, ConfigBase):
            return value.to_dict()
        if isinstance(value, (list, tuple)):
            return [self.serialize_value(item) for item in value]
        if isinstance(value, dict):
            return {k: self.serialize_value(v) for k, v in value.items()}
        return value

    def deserialize_value(self, key, value):
        """Helper method to deserialize different types of values."""
        if isinstance(value, dict):
            if "type" in value and value["type"] == "function":
                try:
                    module = __import__(
                        value["module"], fromlist=[value["function_name"]]
                    )
                    return getattr(module, value["function_name"])
                except (ImportError, AttributeError) as e:
                    raise RuntimeError(
                        f"Cannot import function {value['function_name']} "
                        f"from module {value['module']}"
                    ) from e
            # Check if this is a nested config
            if "_config" in key:
                return self._create_config_instance(key, value)
            # Normal dict
            return {k: self.deserialize_value(k, v) for k, v in value.items()}
        elif isinstance(value, list):
            return [self.deserialize_value(key, v) for v in value]
        elif isinstance(value, str):
            # Try to detect if the string represents a path
            path_indicators = ["/", "\\", "output", "data"]
            if any(indicator in value for indicator in path_indicators):
                try:
                    path = Path(value)
                    # Additional validation could be added here if needed
                    return path
                except Exception:
                    return value
        return value

    def _create_config_instance(self, key, data):
        """Create and populate a config instance based on data structure."""
        # Determine the config class type based on the data
        config_class = None
        if "detector" in key:
            config_class = DetectorConfig
        elif "acquisition" in key:
            config_class = AcquisitionConfig
        elif "phantom" in key:
            config_class = PhantomConfig
        elif "source" in key:
            config_class = SourceConfig
        elif "free_flight" in key:
            config_class = FreeFlightConfig
        elif "garf" in key:
            config_class = GARFConfig

        if config_class:
            instance = config_class(
                spect_config=self
            )  # getattr(self, "spect_config", None))
            instance.from_dict(data)
            return instance
        return data

    def get_excluded_keys(self):
        """Keys to exclude from serialisation (e.g. circular references)."""
        return ["spect_config"] if hasattr(self, "spect_config") else []

    def to_dict(self):
        """Convert configuration to dictionary, handling special types."""
        excluded = self.get_excluded_keys()
        return {
            k: self.serialize_value(v)
            for k, v in vars(self).items()
            if k not in excluded
        }

    def from_dict(self, data):
        """Populate configuration from a dictionary, handling special types."""
        for key, value in data.items():
            if key not in self.get_excluded_keys():
                setattr(self, key, self.deserialize_value(key, value))
        return self

    def to_json(self, file_path=None):
        """Serialise configuration to JSON."""
        json_dict = self.to_dict()
        if file_path:
            with open(file_path, "w") as f:
                json.dump(json_dict, f, indent=4)
        return json.dumps(json_dict, indent=4)

    @classmethod
    def from_json(cls, file_path=None, json_str=None):
        """Create a configuration instance from JSON."""
        if file_path is not None:
            with open(file_path, "r") as f:
                data = json.load(f)
        elif json_str is not None:
            data = json.loads(json_str)
        else:
            raise ValueError("Either file_path or json_str must be provided")

        instance = cls()
        instance.from_dict(data)
        return instance


class SPECTConfig(ConfigBase):
    """
    Represents the configuration object for a SPECT simulation.

    This class organises and manages the configuration details required to set up
    and execute a SPECT (Single Photon Emission Computed Tomography) simulation.

    It includes settings for the simulation output, main configurable elements,
    such as detector, phantom, source, and protocol, as well as simulation
    initialisation and execution methods.
    """

    def __init__(self, simu_name="spect"):
        # default
        self.output_folder = Path("output")
        self.output_basename = "projection.mhd"
        self.simu_name = simu_name
        self.number_of_threads = 1

        # only used if ff
        self.output_folder_primary = None
        self.output_folder_scatter = None

        # main elements
        self.detector_config = DetectorConfig(self)
        self.phantom_config = PhantomConfig(self)
        self.source_config = SourceConfig(self)
        self.acquisition_config = AcquisitionConfig(self)
        self.free_flight_config = FreeFlightConfig(self)

    def __str__(self):
        s = f"SPECT simulation\n"
        s += f"Output folder: {self.output_folder}\n"
        s += f"Output basename: {self.output_basename}\n"
        s += f"Nb of threads: {self.number_of_threads}\n"
        s += f"{self.detector_config}\n"
        s += f"{self.phantom_config}\n"
        s += f"{self.source_config}\n"
        s += f"{self.acquisition_config}"
        return s

    def setup_simulation(self, sim, visu=False):
        # create the output folder if not exist
        os.makedirs(self.output_folder, exist_ok=True)
        # default initialization
        self.initialize_simulation(sim, visu)
        # init all elements
        self.detector_config.setup_simulation(sim)
        phantom = self.phantom_config.setup_simulation(sim)
        self.source_config.setup_simulation(sim, phantom)
        self.acquisition_config.setup_simulation(sim, self.detector_config.head_names)

    def setup_simulation_ff_primary(self, sim, sources=None, visu=False):
        # we temporarily change the output folder
        save_folder = Path(self.output_folder)
        self.output_folder_primary = save_folder / "primary"
        self.output_folder = self.output_folder_primary
        # set the initial simulation
        self.source_config.total_activity = self.free_flight_config.primary_activity
        self.setup_simulation(sim, visu)
        # but keep the initial folder for further use of config
        self.output_folder = save_folder
        if sources is None:
            n = f"{self.simu_name}_source"
            source = sim.source_manager.get_source(n)
            sources = [source]
        for source in sources:
            self.free_flight_config.setup_simulation_primary(sim, source)
        self.dump_ff_info_primary()

    def setup_simulation_ff_scatter(self, sim, sources=None, visu=False):

        # because primary was probably config/run before we clean the
        # static variables from G4 to avoid issues.
        # g4.GateGammaFreeFlightOptrActor.ClearOperators()

        # we temporarily change the output folder
        save_folder = Path(self.output_folder)
        self.output_folder_scatter = save_folder / "scatter"
        self.output_folder = self.output_folder_scatter
        # set the initial simulation
        self.source_config.total_activity = self.free_flight_config.scatter_activity
        self.setup_simulation(sim, visu)
        # but keep the initial folder for further use of config
        self.output_folder = save_folder
        if sources is None:
            name = f"{self.simu_name}_source"
            source = sim.source_manager.get_source(name)
            sources = [source]
        for source in sources:
            self.free_flight_config.setup_simulation_scatter(sim, source)
        self.dump_ff_info_scatter()

    def dump_ff_info_primary(self, filename="ff_info.json"):
        n = {
            "primary_activity": self.free_flight_config.primary_activity / g4_units.Bq,
            "angle_tolerance": self.free_flight_config.angle_tolerance / g4_units.deg,
        }
        fn = self.output_folder_primary / filename
        with open(fn, "w") as f:
            json.dump(n, f, indent=4)

    def dump_ff_info_scatter(self, filename="ff_info.json"):
        n = {
            "scatter_activity": self.free_flight_config.scatter_activity / g4_units.Bq,
            "angle_tolerance": self.free_flight_config.angle_tolerance / g4_units.deg,
            "max_compton_level": self.free_flight_config.max_compton_level,
            "compton_splitting_factor": self.free_flight_config.compton_splitting_factor,
            "rayleigh_splitting_factor": self.free_flight_config.rayleigh_splitting_factor,
        }
        fn = self.output_folder_scatter / filename
        with open(fn, "w") as f:
            json.dump(n, f, indent=4)

    def initialize_simulation(self, sim, visu):
        # main options
        sim.random_seed = "auto"
        sim.check_volumes_overlap = True
        sim.visu = visu
        sim.visu_type = "qt"
        sim.output_dir = self.output_folder
        sim.progress_bar = True
        sim.store_json_archive = True
        sim.store_input_files = False
        sim.json_archive_filename = f"simulation.json"

        # threads
        if visu:
            self.number_of_threads = 1
        sim.number_of_threads = self.number_of_threads

        # world
        m = gate.g4_units.m
        sim.world.size = [3 * m, 3 * m, 3 * m]
        sim.world.material = "G4_AIR"

        # physics
        sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
        sim.physics_manager.global_production_cuts.all = 1 * gate.g4_units.mm

        # add the "stats" actor
        stats = sim.add_actor("SimulationStatisticsActor", "stats")
        stats.output_filename = "stats.txt"
        stats.track_types_flag = True


class PhantomConfig(ConfigBase):
    """
    This class is used in SPECTConfig. Represents a phantom configuration

    This class is used to configure and handle a phantom in a medical imaging
    simulation context. It includes properties for the phantom image,
    labels, density tolerance, and provides methods for printing the
    phantom configuration and generating a simulation. The phantom's
    construction and materials are based on Hounsfield Unit (HU) mappings
    to materials defined in external files.
    """

    def __init__(self, spect_config):
        gcm3 = gate.g4_units.g_cm3
        self.spect_config = spect_config
        # user param
        self.image = None
        self.labels = None
        self.material_db = None
        self.density_tol = 0.05 * gcm3
        self.translation = None

    def __str__(self):
        s = f"Phantom image: {self.image}\n"
        s += f"Phantom labels: {self.labels}"
        return s

    def setup_simulation(self, sim):
        # can be: nothing or voxelized
        if self.image is None:
            return None
        # special case for visu
        if sim.visu is True:
            phantom = self.add_fake_phantom_for_visu(sim)
            if self.translation is not None:
                phantom.translation = self.translation
            return phantom
        # insert voxelized phantom
        phantom = sim.add_volume("Image", f"{self.spect_config.simu_name}_phantom")
        phantom.image = self.image
        phantom.material = "G4_AIR"
        if self.material_db is not None:
            sim.volume_manager.add_material_database(self.material_db)
        if self.labels is not None:
            phantom.read_label_to_material(self.labels)
        else:
            f1 = gate.utility.get_data_folder() / "Schneider2000MaterialsTable.txt"
            f2 = gate.utility.get_data_folder() / "Schneider2000DensitiesTable.txt"
            vm, materials = gate.geometry.materials.HounsfieldUnit_to_material(
                sim, self.density_tol, f1, f2
            )
            phantom.voxel_materials = vm

        # translation ?
        if self.translation is not None:
            phantom.translation = self.translation

        return phantom

    def add_fake_phantom_for_visu(self, sim):
        gate.exception.warning(f"FAKE voxelized phantom for visu: {self.image}")
        img_info = read_image_info(self.image)
        phantom = sim.add_volume("Box", f"{self.spect_config.simu_name}_phantom")
        phantom.material = "G4_WATER"
        phantom.size = img_info.size * img_info.spacing
        return phantom


class DetectorConfig(ConfigBase):
    """
    Used in SPECTConfig. Represents the configuration for a SPECT detector.

    Manage: detector models, crystals, collimators, digitizers,
    and the number of detector heads.
    """

    def __init__(self, spect_config):
        self.spect_config = spect_config
        # spect models
        self.available_models = ["intevo", "nm670"]
        # user param
        self.model = None
        self.collimator = None
        self.digitizer_function = None
        self.digitizer_channels = None
        self.number_of_heads = 2
        self.size = None
        self.spacing = None
        # computed (not user-defined)
        self.head_names = []
        self.proj_names = []
        # fix later
        self.garf_config = GARFConfig(spect_config)

    def __str__(self):
        s = f"Detector model: {self.model}\n"
        s += f"Detector collimator: {self.collimator}\n"
        s += f"Detector digitizer: {self.digitizer_function}\n"
        s += f"Detector # of heads: {self.number_of_heads}"
        return s

    def get_model_module(self):
        m = None
        if self.model == "nm670":
            m = nm670
        if self.model == "intevo":
            m = intevo
        # FIXME veriton
        if m is None:
            fatal(
                f'Unknown detector model: "{self.model}", available models: {self.available_models}'
            )
        return m

    def get_heads(self, sim):
        # only set after initialisation
        detectors = [sim.volume_manager.get_volume(name) for name in self.head_names]
        return detectors

    def get_detector_normal(self):
        if self.model == "nm670":  # FIXME must be in intevo / nm670 module
            return [0, 0, -1]
        if self.model == "intevo":
            return [1, 0, 0]
        fatal(f"Unknown detector model: {self.model}")
        return None

    def get_proj_actors(self, sim):
        # only set after initialisation
        projs = [sim.actor_manager.get_actor(name) for name in self.proj_names]
        return projs

    def get_proj_base_filename(self, i):
        filename, ext = get_basename_and_extension(self.spect_config.output_basename)
        f = f"{filename}_{i}{ext}"
        return f

    def get_proj_filenames(self, sim):
        projs = self.get_proj_actors(sim)
        filenames = [
            self.spect_config.output_folder / proj.get_output_path("counts").name
            for proj in projs
        ]
        return filenames

    def setup_simulation(self, sim):
        if self.model not in self.available_models:
            fatal(
                f'The model "{self.model}" is unknown. '
                f"Known models are: {self.available_models}"
            )

        # get the module intevo or nm670
        m = self.get_model_module()

        # GARF ?
        if self.garf_config.is_enabled():
            if self.size is None:
                self.size = m.get_default_size_and_spacing()[0]
            if self.spacing is None:
                self.spacing = m.get_default_size_and_spacing()[1]
            self.head_names = self.garf_config.create_simulation(sim)
            return

        # digit function ?
        if self.digitizer_function is None:
            self.digitizer_function = m.add_digitizer
        self.head_names = []
        self.proj_names = []

        # channels? Keep the initial digitizer_channels value.
        channels = self.digitizer_channels
        if self.digitizer_channels is None:
            rad = self.spect_config.source_config.radionuclide
            if rad is None:
                fatal(
                    f'Unknown radionuclide: "{self.spect_config.source_config.radionuclide}"'
                )
            channels = get_default_energy_windows(rad)

        # create the SPECT detector for each head
        simu_name = self.spect_config.simu_name
        for i in range(self.number_of_heads):
            # Create the head detector
            hn = f"{simu_name}_spect{i}"
            self.head_names.append(hn)
            det, colli, crystal = m.add_spect_head(
                sim,
                hn,
                collimator_type=self.collimator,
                debug=sim.visu == True,
            )
            # set the digitizer
            dname = f"{simu_name}_digit{i}"
            self.digitizer_function(
                sim,
                crystal.name,
                dname,
                self.size,
                self.spacing,
                channels,
                self.get_proj_base_filename(i),
            )
            proj = sim.actor_manager.find_actor_by_type(
                "DigitizerProjectionActor", dname
            )
            # update size/spacing that may have been set in the digitizer_function
            self.size = proj.size
            self.spacing = proj.spacing
            # keep the name of the projection actors
            self.proj_names.append(proj.name)


class GARFConfig(ConfigBase):
    """
    Used in DetectorConfig.
    Configuration class for GARF detector.
    """

    def __init__(self, spect_config):
        self.spect_config = spect_config
        # options
        self.pth_filename = None
        self.batch_size = 1e5
        self.verbose_batch = True
        self.gpu_mode = "auto"
        self.arf_actors = []

    def __str__(self):
        s = f"GARF : {self.pth_filename}\n"
        s += f"GARF image size: {self.spect_config.detector_config.size}\n"
        s += f"GARF image spacing: {self.spect_config.detector_config.spacing}\n"
        s += f"GARF batch size: {self.batch_size}"
        return s

    def is_enabled(self):
        return self.pth_filename is not None

    def create_simulation(self, sim):
        # get the module (intevo, nm670)
        m = self.spect_config.detector_config.get_model_module()
        colli = self.spect_config.detector_config.collimator
        size = self.spect_config.detector_config.size
        spacing = self.spect_config.detector_config.spacing
        head_names = []
        for i in range(self.spect_config.detector_config.number_of_heads):
            name = f"{self.spect_config.simu_name}__arf_{i}"
            _, arf = m.add_arf_detector(
                sim, name, colli, size, spacing, self.pth_filename
            )
            # same name as conventional detector
            arf.output_filename = (
                self.spect_config.detector_config.get_proj_base_filename(i)
            )
            self.arf_actors.append(arf)
            head_names.append(name)
        return head_names


class SourceConfig(ConfigBase):
    """
    Used in SPECTConfig. Represents a configuration for an activity source.

    This class is used to define and manage the settings of an activity source,
    such as the source image, radionuclide, and total activity.
    """

    def __init__(self, spect_config):
        self.spect_config = spect_config
        # user param
        self.image = None
        self.radionuclide = None
        self.total_activity = 1 * g4_units.Bq
        # self.gaga = GAGAConfig(spect_config) # FIXME todo

    def __str__(self):
        s = f"Activity source image: {self.image}\n"
        s += f"Activity source radionuclide: {self.radionuclide}\n"
        s += f"Activity source: {self.total_activity / gate.g4_units.Bq} Bq"
        return s

    def setup_simulation(self, sim, phantom):
        # can be: nothing or voxelized or gaga (later)
        if self.image is None:
            return
        if self.radionuclide is None:
            fatal(f"Radionuclide is None, please set a radionuclide (eg. 'lu177')")

        # set the source
        source = sim.add_source("VoxelSource", f"{self.spect_config.simu_name}_source")
        source.attached_to = phantom
        source.image = str(self.image)
        # mode voxelized source according to voxelized phantom
        if self.spect_config.phantom_config.image is not None:
            source.position.translation = (
                gate.image.get_translation_between_images_center(
                    self.spect_config.phantom_config.image, source.image
                )
            )
        set_source_energy_spectrum(source, self.radionuclide)
        source.particle = "gamma"
        source.activity = self.total_activity / self.spect_config.number_of_threads
        if sim.visu is True:
            source.activity = 10 * gate.g4_units.Bq


class AcquisitionConfig(ConfigBase):
    """
    Used in SPECTConfig. A configuration class for acquisition settings specific to a simulation.

    This class provides configuration parameters and functionality for managing
    acquisition settings in SPECT simulations. It allows specifying parameters
    such as acquisition duration, radius, and number of angles.
    """

    def __init__(self, spect_config):
        self.spect_config = spect_config
        # user param
        self.duration = 1 * gate.g4_units.s
        self.radius = 30 * g4_units.cm
        # self.angles = None
        self.number_of_angles = 1
        # internal var
        self.available_starting_head_angles = {
            "1": [0.0],
            "2": [0.0, 180.0],
            "3": [0.0, 120.0, 240.0],
            "4": [0.0, 90.0, 180.0, 270.0],
            # FIXME later for Veriton
        }

    def __str__(self):
        s = f"Acquisition duration: {self.duration / g4_units.s} sec\n"
        s += f"Acquisition radius: {self.radius / g4_units.mm} mm\n"
        s += f"Acquisition # angles: {self.number_of_angles}"
        return s

    def setup_simulation(self, sim, detector_names):
        # get the number of heads and starting angles
        nb = str(self.spect_config.detector_config.number_of_heads)
        if nb not in self.available_starting_head_angles:
            fatal(f"The number of heads must in {self.available_starting_head_angles}")
        starting_head_angles = self.available_starting_head_angles[nb]

        # set the rotation angles (runs)
        step_time = self.duration / self.number_of_angles
        sim.run_timing_intervals = [
            [i * step_time, (i + 1) * step_time] for i in range(self.number_of_angles)
        ]

        # get the detectors
        detectors = [sim.volume_manager.get_volume(n) for n in detector_names]

        # compute the gantry rotations
        step_angle = 360.0 / len(starting_head_angles) / self.number_of_angles
        m = self.spect_config.detector_config.get_model_module()
        i = 0
        for sa in starting_head_angles:
            m.rotate_gantry(
                detectors[i], self.radius, sa, step_angle, self.number_of_angles
            )
            i += 1


class FreeFlightConfig(ConfigBase):

    def __init__(self, spect_config):
        self.spect_config = spect_config
        # user param
        self.angle_tolerance = 6 * g4_units.deg
        self.forced_direction_flag = True
        self.angle_tolerance_min_distance = 6 * g4_units.cm
        self.max_compton_level = 5
        self.compton_splitting_factor = 50
        self.rayleigh_splitting_factor = 50
        self.minimal_weight = 1e-30
        self.primary_activity = 1 * g4_units.Bq
        self.scatter_activity = 1 * g4_units.Bq
        self.max_rejection = None
        # primary: do not bias in those volumes
        self.primary_unbiased_volumes = "detector"
        # scatter options
        self.scatter_unbiased_volumes = "detector"
        self.scatter_kill_interacting_in_volumes = "crystal"

    def __str__(self):
        s = f"FreeFlight FD: {self.forced_direction_flag}\n"
        s += f"FreeFlight angle tol: {self.angle_tolerance / g4_units.deg} deg\n"
        s += f"FreeFlight angle tol min dist: {self.angle_tolerance_min_distance / g4_units.mm} mm\n"
        s += f"FreeFlight max_compton_level: {self.max_compton_level}\n"
        s += f"FreeFlight compton_splitting_factor: {self.compton_splitting_factor}\n"
        s += f"FreeFlight rayleigh_splitting_factor: {self.rayleigh_splitting_factor}\n"
        s += f"FreeFlight minimal_weight: {self.minimal_weight}\n"
        s += f"FreeFlight primary_activity: {self.primary_activity / g4_units.Bq} Bq\n"
        s += f"FreeFlight scatter_activity: {self.scatter_activity / g4_units.Bq} Bq\n"
        return s

    def initialize(self, sim):
        # Weights MUST be in the digitizer
        # FIXME change the way to get the hits
        hits_actors = sim.actor_manager.find_actors("_hits")
        if len(hits_actors) != 0:
            for ha in hits_actors:
                if "Weight" not in ha.attributes:
                    ha.attributes.append("Weight")
        # be sure the squared counts flag is enabled
        proj_actors = sim.actor_manager.find_actors("projection")
        if len(proj_actors) != 0:
            for d in proj_actors:
                d.squared_counts.active = True

        # if garf is used, compute the squared counts
        if self.spect_config.detector_config.garf_config.pth_filename is not None:
            for arf in self.spect_config.detector_config.garf_config.arf_actors:
                arf.squared_counts.active = True

        # GeneralProcess must *NOT* be used
        s = f"/process/em/UseGeneralProcess false"
        if s not in sim.g4_commands_before_init:
            sim.g4_commands_before_init.append(s)

    def get_crystal_volume_names(self):
        volume_names = [
            f"{self.spect_config.simu_name}_spect{i}_crystal"
            for i in range(self.spect_config.detector_config.number_of_heads)
        ]
        return volume_names

    def get_detector_volume_names(self):
        volume_names = self.spect_config.detector_config.head_names
        return volume_names

    def setup_simulation_primary(self, sim, source):
        self.initialize(sim)

        # consider the volume where we stop applying ff
        target_volume_names = None
        if self.primary_unbiased_volumes == "crystal":
            target_volume_names = self.get_crystal_volume_names()
        elif self.primary_unbiased_volumes == "detector":
            target_volume_names = self.get_detector_volume_names()
        else:
            fatal(f"Unknown ff ignored volume: {self.primary_unbiased_volumes}")

        # add the ff actor (only once !)
        ff_name = f"{self.spect_config.simu_name}_ff"
        if ff_name not in sim.actor_manager.actors:
            ff = sim.add_actor("GammaFreeFlightActor", ff_name)
            ff.attached_to = "world"
            ff.unbiased_volumes = target_volume_names
            ff.minimal_weight = self.minimal_weight
        else:
            ff = sim.actor_manager.get_actor(ff_name)

        if self.forced_direction_flag:
            self.setup_forced_detection(sim, source, target_volume_names)
        else:
            self.setup_acceptance_angle(sim, source, target_volume_names)

        return ff

    def setup_acceptance_angle(self, sim, source, target_volume_names):
        detector_config = self.spect_config.detector_config
        normal_vector = detector_config.get_detector_normal()
        n = self.spect_config.number_of_threads
        source.activity = self.primary_activity / n
        source.direction.acceptance_angle.forced_direction_flag = False
        source.direction.acceptance_angle.skip_policy = "SkipEvents"
        if self.max_rejection is not None:
            source.direction.acceptance_angle.max_rejection = self.max_rejection
        source.direction.acceptance_angle.volumes = target_volume_names
        source.direction.acceptance_angle.intersection_flag = True
        source.direction.acceptance_angle.normal_flag = True
        source.direction.acceptance_angle.normal_vector = normal_vector
        source.direction.acceptance_angle.normal_tolerance = self.angle_tolerance
        # distance dependent normal tol is very slow, dont use it
        source.direction.acceptance_angle.distance_dependent_normal_tolerance = False
        # minimal distance should not be used for primary
        source.direction.acceptance_angle.normal_tolerance_min_distance = 0

        return source

    def setup_forced_detection(self, sim, source, target_volume_names):
        detector_config = self.spect_config.detector_config

        # need an additional source copy for each head
        sources = [source]
        for i in range(1, detector_config.number_of_heads):
            s = sim.source_manager.add_source_copy(source.name, f"{source.name}_{i}")
            sources.append(s)

        normal_vector = detector_config.get_detector_normal()
        i = 0

        for source in sources:
            # force the direction to one single volume for each source
            source.direction.acceptance_angle.volumes = [target_volume_names[i]]
            source.direction.acceptance_angle.normal_vector = normal_vector
            source.direction.acceptance_angle.normal_tolerance = self.angle_tolerance
            source.direction.acceptance_angle.skip_policy = "SkipEvents"
            source.direction.acceptance_angle.normal_flag = False
            source.direction.acceptance_angle.intersection_flag = False
            source.direction.acceptance_angle.forced_direction_flag = True
            i += 1
        return sources

    def setup_simulation_scatter(self, sim, source):
        self.initialize(sim)

        target_volume_names = None
        if self.scatter_unbiased_volumes == "crystal":
            target_volume_names = self.get_crystal_volume_names()
        elif self.scatter_unbiased_volumes == "detector":
            target_volume_names = self.get_detector_volume_names()
        else:
            fatal(
                f"Unknown ff-scatter unbiased volume: {self.scatter_unbiased_volumes}"
            )

        kill_volumes = []
        if self.scatter_kill_interacting_in_volumes == "crystal":
            kill_volumes = self.get_crystal_volume_names()
        elif self.scatter_kill_interacting_in_volumes == "detector":
            kill_volumes = self.get_detector_volume_names()

        print("unbiased_volumes:", target_volume_names)
        print("kill_volumes:", kill_volumes)

        n = self.spect_config.number_of_threads
        source.activity = self.scatter_activity / n

        # set the FF actor for scatter
        normal_vector = self.spect_config.detector_config.get_detector_normal()
        g4.GateGammaFreeFlightOptrActor.ClearOperators()  # needed linux when no MT ?
        ff = sim.add_actor(
            "ScatterSplittingFreeFlightActor",
            f"{self.spect_config.simu_name}_ff",
        )
        ff.attached_to = "world"  # FIXME -> remove this, always world + ignored_vol ?
        ff.minimal_weight = self.minimal_weight
        ff.unbiased_volumes = target_volume_names
        ff.kill_interacting_in_volumes = kill_volumes
        ff.compton_splitting_factor = self.compton_splitting_factor
        ff.rayleigh_splitting_factor = self.rayleigh_splitting_factor
        ff.max_compton_level = self.max_compton_level
        ff.acceptance_angle.intersection_flag = True
        ff.acceptance_angle.normal_flag = True
        ff.acceptance_angle.forced_direction_flag = False
        ff.acceptance_angle.volumes = target_volume_names
        ff.acceptance_angle.normal_vector = normal_vector
        ff.acceptance_angle.normal_tolerance = self.angle_tolerance
        ff.acceptance_angle.normal_tolerance_min_distance = (
            self.angle_tolerance_min_distance
        )
        ff.acceptance_angle.distance_dependent_normal_tolerance = False

        """ REMOVED because too slow
        ff.acceptance_angle.distance_dependent_normal_tolerance = True
        tol = options.scatter_angle_tolerance
        ff.acceptance_angle.distance1 = tol[0]
        ff.acceptance_angle.angle1 = tol[1]
        ff.acceptance_angle.distance2 = tol[2]
        ff.acceptance_angle.angle2 = tol[3]
        """

        return ff


def spect_freeflight_merge_all_heads(
    folder,
    n_prim,
    n_scatter,
    n_target,
    prim_folder="primary",
    scatter_folder="scatter",
    nb_of_heads=2,
    counts_filename_pattern="projection_$I_counts.mhd",
    sq_counts_filename_pattern="projection_$I_squared_counts.mhd",
    mean_filename="projection_$I_counts.mhd",
    rel_uncert_suffix="relative_uncertainty_$I",
    verbose=True,
):
    for d in range(nb_of_heads):
        spect_freeflight_merge(
            folder,
            n_prim,
            n_scatter,
            n_target,
            prim_folder=prim_folder,
            scatter_folder=scatter_folder,
            counts_filename=counts_filename_pattern.replace("$I", str(d)),
            sq_counts_filename=sq_counts_filename_pattern.replace("$I", str(d)),
            mean_filename=mean_filename.replace("$I", str(d)),
            rel_uncert_suffix=rel_uncert_suffix.replace("$I", str(d)),
            verbose=verbose,
        )


def spect_freeflight_merge(
    folder,
    n_prim,
    n_scatter,
    n_target,
    prim_folder="primary",
    scatter_folder="scatter",
    counts_filename="projection_0_counts.mhd",
    sq_counts_filename="projection_0_squared_counts.mhd",
    mean_filename="mean.mhd",
    rel_uncert_suffix="relative_uncertainty",
    verbose=True,
):
    # make them path
    prim_folder = Path(prim_folder)
    scatter_folder = Path(scatter_folder)

    # primary
    if n_prim > 0:
        img = folder / prim_folder / counts_filename
        sq_img = folder / prim_folder / sq_counts_filename
        out = folder / prim_folder / f"{img.stem}_{rel_uncert_suffix}.mhd"
        _, prim, prim_squared = history_rel_uncertainty_from_files(
            img, sq_img, n_prim, out
        )
    else:
        prim = None
        prim_squared = None

    # scatter
    if n_scatter > 0:
        img = folder / scatter_folder / counts_filename
        sq_img = folder / scatter_folder / sq_counts_filename
        out = folder / scatter_folder / f"{img.stem}_{rel_uncert_suffix}.mhd"
        _, scatter, scatter_squared = history_rel_uncertainty_from_files(
            img, sq_img, n_scatter, out
        )
    else:
        scatter = None
        scatter_squared = None

    # combined
    uncert, mean = history_ff_combined_rel_uncertainty(
        prim, prim_squared, scatter, scatter_squared, n_prim, n_scatter
    )

    scaling = n_target / n_prim
    mean = mean * scaling
    if verbose:
        print(f"Primary n = {n_prim}  Scatter n = {n_scatter}  Target n = {n_target}")
        if n_scatter > 0:
            print(f"Primary to scatter ratio = {n_prim / n_scatter:.01f}")
        print(f"Scaling to target        = {scaling:.01f}")

    # write combined image
    prim_img = sitk.ReadImage(img)
    img = sitk.GetImageFromArray(mean)
    img.CopyInformation(prim_img)
    fn = folder / mean_filename
    sitk.WriteImage(img, fn)
    if verbose:
        print(fn)

    # write combined relative uncertainty
    img = sitk.GetImageFromArray(uncert)
    img.CopyInformation(prim_img)
    fn = folder / f"{fn.stem}_{rel_uncert_suffix}.mhd"
    sitk.WriteImage(img, fn)
    if verbose:
        print(fn)

    # open info if the file exists
    prim_info = {}
    if n_prim > 0:
        prim_info_fn = folder / prim_folder / "ff_info.json"
        if prim_info_fn.is_file():
            with open(prim_info_fn, "r") as f:
                prim_info = json.load(f)

    # open info if the file exists
    scatter_info = {}
    if n_scatter > 0:
        scatter_info_fn = folder / scatter_folder / "ff_info.json"
        if scatter_info_fn.is_file():
            with open(scatter_info_fn, "r") as f:
                scatter_info = json.load(f)

    # write combined information
    info = prim_info
    info.update(scatter_info)
    info_fn = folder / "ff_info.json"
    with open(info_fn, "w") as f:
        json.dump(info, f, indent=4)


def plot_ddaa(acceptance_angle, output_filename=None):
    # plot dd
    a1 = acceptance_angle.angle1
    a2 = acceptance_angle.angle2
    d1 = acceptance_angle.distance1
    d2 = acceptance_angle.distance2
    distances = np.linspace(d1 / 2, d2 * 2, 200)
    angles = [distance_dependent_angle_tolerance(a1, a2, d1, d2, d) for d in distances]

    cm = g4_units.cm
    plt.figure(figsize=(8, 6))
    plt.plot(distances / cm, np.degrees(angles), label="Distance vs Angle")
    plt.xlabel("Distance (cm)")
    plt.ylabel("Angle (degrees)")
    plt.title("Distance vs Angle Tolerance")
    plt.grid()
    plt.legend()
    # plt.show()
    if output_filename is not None:
        plt.savefig(output_filename)
    return plt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import uproot
import opengate as gate
import numpy as np
from collections import defaultdict
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt


#########################################################################################
# Constants
#########################################################################################
# Units
m = gate.g4_units.m
cm = gate.g4_units.cm
mm = gate.g4_units.mm
eV = gate.g4_units.eV
MeV = gate.g4_units.MeV
Bq = gate.g4_units.Bq
gcm3 = gate.g4_units.g_cm3
deg = gate.g4_units.deg


#########################################################################################
# Recurring elements of the tests in terms of simulation
#########################################################################################
def setup_simulation_engine(path):
    # create simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.random_seed = 12345654
    sim.progress_bar = True

    # add a material database
    sim.volume_manager.add_material_database(path.data / "GateMaterials.db")

    # set the world size
    sim.world.size = [3 * m, 3 * m, 3 * m]
    sim.world.material = "G4_AIR"

    # set the physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.set_production_cut("waterbox", "gamma", 10 * mm)
    sim.physics_manager.enable_decay = True

    return sim


def setup_actor(sim, actor_name, volume_name):
    # add phase actor
    curr_actor = sim.add_actor("PhaseSpaceActor", actor_name)
    curr_actor.attached_to = volume_name
    curr_actor.attributes = [
        "EventID",
        "TrackID",
        "PrePosition",
        "PreDirection",
        "PostDirection",
        "ParticleName",
        "TrackCreatorProcess",
        "KineticEnergy",
    ]
    curr_actor.steps_to_store = "first"
    f = sim.add_filter("ParticleFilter", f"f_{actor_name}")
    f.particle = "gamma"
    curr_actor.filters.append(f)

    return curr_actor


#########################################################################################
# Methods used to ascertain the test behavior
#########################################################################################
def calculate_angle(dir1, dir2):
    # Calculate the dot product
    dot_product = np.dot(dir1, dir2)

    # Calculate magnitudes
    magnitude1 = np.linalg.norm(dir1)
    magnitude2 = np.linalg.norm(dir2)

    # Calculate the angle between the directions
    # cos_theta = dot_product / (magnitude1 * magnitude2)
    cos_theta = np.divide(
        dot_product,
        (magnitude1 * magnitude2),
        out=np.zeros_like(dot_product),
        where=(magnitude1 * magnitude2) != 0,
    )
    return np.arccos(np.clip(cos_theta, -1.0, 1.0))  # Clip to avoid numerical errors


def read_gamma_pairs(root_filename, actor_name="phsp", is_btb=False):
    # Load the ROOT file
    file = uproot.open(root_filename)

    # Access the tree
    tree = file[actor_name]

    # Extract relevant branches
    event_id = tree["EventID"].array(library="np")
    particle_name = tree["ParticleName"].array(library="np")
    creator_process = tree["TrackCreatorProcess"].array(library="np")

    # Extract PreDirection components
    pre_dir_x = tree["PreDirection_X"].array(library="np")
    pre_dir_y = tree["PreDirection_Y"].array(library="np")
    pre_dir_z = tree["PreDirection_Z"].array(library="np")

    # Combine components into a single 3D vector for PreDirection
    pre_dir = np.vstack((pre_dir_x, pre_dir_y, pre_dir_z)).T

    if is_btb:
        gamma_mask = particle_name == "gamma"
    else:
        # Filter for gamma particles created by the annihilation process
        gamma_mask = (particle_name == "gamma") & (creator_process == "annihil")

    # Dictionary to store paired gammas by EventID
    event_id = event_id[gamma_mask]
    pre_dir = pre_dir[gamma_mask]
    event_dict = defaultdict(list)

    # Grouping gamma particles by EventID
    for i, eid in enumerate(event_id):
        event_dict[eid].append(pre_dir[i])

    # Filter out events with exactly 2 gammas
    gamma_pairs = [dirs for dirs in event_dict.values() if len(dirs) == 2]

    return gamma_pairs


def compute_acollinearity_angles(gamma_pairs):
    acollinearity_angles = []

    for pair in gamma_pairs:
        dir1, dir2 = pair
        angle = calculate_angle(dir1, dir2)
        acollinearity = np.abs(
            np.pi - angle
        )  # Acollinearity deviation from 180 degrees
        acollinearity_angles.append(np.degrees(acollinearity))  # Convert to degrees
        if np.degrees(acollinearity) > 20:
            print(
                f"dir1 = {dir1}, dir2 = {dir2} -> angle = {np.degrees(acollinearity)} {np.degrees(angle)}"
            )

    return acollinearity_angles


def rayleigh(abs_value, amp, scale):
    return amp * (abs_value / scale**2) * np.exp(-((abs_value) ** 2) / (2.0 * scale**2))


def fit_rayleigh(hist_data):
    hist_pos = hist_data[1][:-1] + np.diff(hist_data[1])[0]
    init_param = [1.0, 0.5 / 2.355]
    popt, _ = curve_fit(rayleigh, hist_pos, hist_data[0], p0=init_param)

    return popt[0], popt[1]


def plot_colin_case(acollinearity_angles):
    colin_median = np.median(acollinearity_angles)
    print(
        f"median angle: {colin_median}  min={np.min(acollinearity_angles)}   max={np.max(acollinearity_angles)}"
    )

    plt.hist(
        acollinearity_angles,
        bins=71,
        range=(0, 1.0),
        alpha=0.7,
        color="blue",
        label="Default",
    )

    return colin_median


def plot_acolin_case_mepip(ion_pair_mean_energy, acollinearity_angles, is_second=False):
    label = f"With mean energy per Ion par of {ion_pair_mean_energy / eV:.1f} eV"
    scale = plot_acolin_case(label, acollinearity_angles, is_second)

    return scale


def plot_acolin_case_angle(acolin_FWHM, acollinearity_angles, is_second=False):
    label = f"Acolin. set at {acolin_FWHM / deg:.2f} deg FWHM"
    scale = plot_acolin_case(label, acollinearity_angles, is_second)

    return scale


def plot_acolin_case(label, acollinearity_angles, is_second=False):
    # If two cases of acolin is shown on the same figure, make sure that we can tell
    # them apart
    if is_second:
        hist_color = "brown"
    else:
        hist_color = "red"

    # Range of 0.0 to 1.0 is enforced since in some rare instances, acolinearity is
    # very large, which skew the histogram.
    data = plt.hist(
        acollinearity_angles,
        bins=71,
        range=(0, 1.0),
        alpha=0.7,
        color=hist_color,
        label=label,
    )

    if is_second:
        box_pos = 0.5 * max(data[0])
        fit_color = "c"
    else:
        plt.ylim((0.0, 2.0 * max(data[0])))
        plt.xlim((0.0, 1.0))
        plt.xlabel("Acollinearity Angle (Degrees)")
        plt.ylabel("Counts")
        plt.title("Acollinearity Distribution of Gamma Pairs")
        plt.grid(True)
        box_pos = max(data[0])
        fit_color = "g"
    plt.legend()

    amplitude, scale = fit_rayleigh(data)
    # Negative value change nothing for the fit but it should be positive.
    scale = np.abs(scale)
    x_value = np.linspace(0.0, 1.0, 50)
    # The norm of a isotropic 2D Gaussian centered at [0.0 0.0] is a Rayleigh
    # distribution with a scale equal to the sigme of the 2D Gaussian.
    plt.plot(x_value, rayleigh(x_value, amplitude, scale), fit_color, linewidth=3)

    textstr = f"FWHM = {2.355 * scale:.2f}°\n$\\sigma$ = {scale:.2f}°"
    props = dict(
        boxstyle="round",
        facecolor="wheat",
        edgecolor=fit_color,
        linewidth=3,
        alpha=0.95,
    )

    plt.text(0.7, box_pos, textstr, bbox=props)

    print(label)
    print(textstr)

    return scale



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test019_linac_phsp_helpers as test019
from opengate.tests import utility

if __name__ == "__main__":
    # create sim
    sim = gate.Simulation()
    test019.create_simu_test019_phsp_source(sim, "_phsp_src")

    # start simulation
    sim.run()

    # print results
    stats = sim.get_actor("Stats")
    print(stats)

    # analyse
    is_ok = test019.analyse_test019_phsp_source(sim)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.contrib.carm.siemensciosalpha import Ciosalpha
from scipy.spatial.transform import Rotation


if __name__ == "__main__":
    # paths
    # paths = utility.get_default_test_paths(__file__, output_folder="test075_siemens_cios_alpha")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.number_of_threads = 1
    sim.random_seed = 12345678
    sim.check_volumes_overlap = True

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    deg = gate.g4_units.deg

    # world
    world = sim.world
    world.size = [5 * m, 5 * m, 5 * m]
    world.material = "G4_AIR"

    # xray tube spectrum parameters
    # tube potential [kV]
    kvp = 100

    # add a carm
    carm = Ciosalpha(sim, kvp, source_only=False)
    carm.rotation = Rotation.from_euler("ZYX", [0, 20, 0], degrees=True).as_matrix()
    carm.translation = [0 * cm, 0 * cm, 0 * cm]
    carm.collimation = [25 * mm, 25 * mm]

    carm.source.n = 1e6
    if sim.visu:
        carm.source.n = 1000

    # aluminum table
    table = sim.add_volume("Box", "aluminum_table")
    table.size = [60 * cm, 2 * m, 0.9 * mm]
    table.material = "G4_Al"
    table.translation = [0 * m, 0 * cm, 0 * cm]
    table.color = [0.8, 0.8, 0.8, 1]

    # start simulation
    sim.run()

    # TODO: Test



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test049_pet_digit_blurring_helpers as t49
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test049_pet_blur", "test049")

    """
    see https://github.com/teaghan/PET_MonteCarlo
    and https://doi.org/10.1002/mp.16032

    PET simulation to test blurring options of the digitizer

    - PET:
    - phantom: nema necr
    - output: singles with and without various blur options
    """

    # create the simulation
    sim = gate.Simulation()
    nb_threads = 2
    t49.create_simulation(sim, nb_threads, fname_suffix="_v2mt")

    # start simulation
    sim.run()

    # print results
    stats = sim.get_actor("Stats")
    print(stats)

    # ----------------------------------------------------------------------------------------------------------
    readout = sim.get_actor("Singles")
    ig = readout.GetIgnoredHitsCount()
    print()
    print(f"Nb of ignored hits : {ig}")

    # check stats
    print()
    gate.exception.warning(f"Check stats")
    p = paths.gate_output
    stats_ref = utility.read_stat_file(p / "stats.txt")
    stats_ref.counts.runs = nb_threads
    is_ok = utility.assert_stats(stats, stats_ref, 0.025)

    # check root hits
    hc = sim.get_actor("Hits")
    f = p / "pet.root"
    is_ok = (
        t49.check_root_hits(paths, 1, f, hc.get_output_path(), "test049_hits_v2_MT.png")
        and is_ok
    )

    # check root singles
    sc = sim.get_actor("Singles")
    is_ok = (
        t49.check_root_singles(
            paths, 1, f, sc.get_output_path(), png_output="test049_singles_v2_MT.png"
        )
        and is_ok
    )

    # gate.delete_run_manager_if_needed(sim) # no
    utility.test_ok(is_ok)



import json
import numpy as np
from pathlib import Path, PurePath

from .exception import fatal

import opengate_core as g4


class GateJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return {
                "__ndarray__": obj.tolist(),
                "__dtype__": str(obj.dtype),
                "__shape__": obj.shape,
            }
        elif isinstance(obj, Path):
            return {"__pathlib_path__": PurePath(obj).parts}
        elif isinstance(obj, g4.G4BestUnit):
            return str(obj).split()
        elif hasattr(obj, "to_dictionary"):
            fatal(
                f"Implementation error: Serializer found GateObject named {obj.name}. "
                f"This should have been turned into a plain dictionary at this stage. "
            )
        else:
            return super().default(obj)


def json_obj_hook(input):
    """
    Decodes a previously encoded numpy ndarray
    with proper shape and dtype
    :param input: (dict) json encoded ndarray
    :return: (ndarray) if input was an encoded ndarray
    """
    if isinstance(input, dict) and "__ndarray__" in input:
        obj = np.array(input["__ndarray__"], input["__dtype__"]).reshape(
            input["__shape__"]
        )
    elif isinstance(input, dict) and "__pathlib_path__" in input:
        obj = Path(input["__pathlib_path__"][0])
        for p in input["__pathlib_path__"][1:]:
            obj /= p
    else:
        obj = input
    return obj


# Overload dump/load from json
def dumps_json(*args, **kwargs):
    kwargs.setdefault("cls", GateJSONEncoder)
    kwargs.setdefault("indent", 4)
    return json.dumps(*args, **kwargs)


def loads_json(*args, **kwargs):
    kwargs.setdefault("object_hook", json_obj_hook)
    return json.loads(*args, **kwargs)


def dump_json(*args, **kwargs):
    kwargs.setdefault("cls", GateJSONEncoder)
    kwargs.setdefault("indent", 4)
    return json.dump(*args, **kwargs)


def load_json(*args, **kwargs):
    kwargs.setdefault("object_hook", json_obj_hook)
    return json.load(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
from opengate.tests import utility
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test044_pbs", "test065_dose2water_from_edep2water_ct"
    )
    output_path = paths.output
    ref_path = paths.output_ref / ".." / "test059"

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12365478910
    sim.random_engine = "MersenneTwister"
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    gcm3 = gate.g4_units.g_cm3

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    ## Beamline model
    IR2HBL = BeamlineModel()
    IR2HBL.name = None
    IR2HBL.radiation_types = "ion 6 12"
    # Nozzle entrance to Isocenter distance
    IR2HBL.distance_nozzle_iso = 1300.00  # 1648 * mm#1300 * mm
    # SMX to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_x = 6700.00
    # SMY to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_y = 7420.00
    # polinomial coefficients
    IR2HBL.energy_mean_coeffs = [11.91893485094217, -9.539517997860457]
    IR2HBL.energy_spread_coeffs = [0.0004790681841295621, 5.253257865904452]
    IR2HBL.sigma_x_coeffs = [2.3335753978880014]
    IR2HBL.theta_x_coeffs = [0.0002944903217664001]
    IR2HBL.epsilon_x_coeffs = [0.0007872786903040108]
    IR2HBL.sigma_y_coeffs = [1.9643343053823967]
    IR2HBL.theta_y_coeffs = [0.0007911780133478402]
    IR2HBL.epsilon_y_coeffs = [0.0024916149017600447]

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # source and beamline info
    beam_data_dict = spots_info_from_txt(
        ref_path / "TreatmentPlan4Gate-F5x5cm_E120MeVn.txt", "ion 6 12", beam_nr=1
    )
    gantry_angle = beam_data_dict["gantry_angle"]

    # nozzle box
    box = sim.add_volume("Box", "box")
    box.size = [500 * mm, 500 * mm, 1000 * mm]
    box.rotation = (
        Rotation.from_euler("z", gantry_angle, degrees=True)
        * Rotation.from_euler("x", -90, degrees=True)
    ).as_matrix()
    if gantry_angle == 0:
        box.translation = [0 * mm, -1148 * mm, 0 * mm]  # [1148 *mm, 0 * mm, 0 * mm]
    elif gantry_angle == 90:
        box.translation = [1148 * mm, 0 * mm, 0 * mm]
    box.material = "Vacuum"
    box.color = [0, 0, 1, 1]

    # nozzle WET
    nozzle = sim.add_volume("Box", "nozzle")
    nozzle.mother = box
    nozzle.size = [500 * mm, 500 * mm, 2 * mm]
    nozzle.material = "G4_WATER"

    # lookup tables
    hu_material = paths.data / "Schneider2000MaterialsTable.txt"
    hu_density = paths.data / "Schneider2000DensitiesTable.txt"

    # ct image
    mhd_ct_path = str(ref_path / "random_HU.mhd")

    # patient
    patient = sim.add_volume("Image", "patient")
    patient.image = mhd_ct_path
    # patient.mother = phantom.name
    # patient.translation = list((img_origin - origin_when_centered) - iso)
    patient.material = "G4_AIR"  # material used by default
    # patient.voxel_materials = [
    #     [-1024, -300, "G4_AIR"],
    #     [-300, 3000, "G4_WATER"],
    # ]
    sim.physics_manager.set_max_step_size(patient.name, 0.8)

    tol = 0.05 * gcm3
    (
        patient.voxel_materials,
        materials,
    ) = gate.geometry.materials.HounsfieldUnit_to_material(
        sim, tol, hu_material, hu_density
    )

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.set_production_cut("world", "all", 1000 * km)

    # add dose actor
    dose_postprocess = sim.add_actor("DoseActor", "dose_postprocess")
    dose_postprocess.dose.output_filename = "dose2water_ct_post.mhd"
    dose_postprocess.attached_to = patient
    dose_postprocess.size = [55, 63, 63]
    dose_postprocess.spacing = [1 * mm, 1 * mm, 1 * mm]
    dose_postprocess.hit_type = "random"
    dose_postprocess.dose.active = True
    dose_postprocess.score_in = "G4_WATER"
    # OPTION CURRENTLY NOT AVAILABLE
    # dose_postprocess.dose_calc_on_the_fly = (
    #     False  # calc dose as edep/mass after end of simulation
    # )

    dose_in_step = sim.add_actor("DoseActor", "dose_in_step")
    dose_in_step.dose.output_filename = "dose2water_ct_step.mhd"
    dose_in_step.attached_to = patient
    dose_in_step.size = [55, 63, 63]
    dose_in_step.spacing = [1 * mm, 1 * mm, 1 * mm]
    dose_in_step.hit_type = "random"
    dose_in_step.dose.active = True
    dose_in_step.score_in = "G4_WATER"
    # CURRENTLY DEACTIVATED OPTION
    # dose_in_step.dose_calc_on_the_fly = (
    #     False  # calculate dose directly in stepping action
    # )

    ## source
    nSim = 4000  # 328935  # particles to simulate per beam
    tps = sim.add_source("TreatmentPlanPBSource", "TP source")
    tps.beam_model = IR2HBL
    tps.n = nSim
    tps.beam_data_dict = beam_data_dict
    tps.beam_nr = 1
    tps.particle = "ion 6 12"

    # start simulation
    run_simulation = True
    if run_simulation:
        # add stat actor
        stat = sim.add_actor("SimulationStatisticsActor", "Stats")
        stat.track_types_flag = True
        # start simulation
        sim.run()

        # print results at the end
        print(stat)

    # read output
    d_post_path = dose_postprocess.dose.get_output_path()
    d_step_path = dose_in_step.dose.get_output_path()
    # img_mhd_out = itk.imread(d_post_path)
    # img_mhd_ref = itk.imread(d_step_path)

    ok = utility.assert_images(d_step_path, d_post_path, tolerance=10)

    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from scipy.spatial.transform import Rotation
import pathlib


def user_hook_volume(simulation_engine):
    sphere_volume = simulation_engine.volume_engine.get_volume("mysphere")
    (
        global_translation_old,
        global_rotation_old,
    ) = gate.geometry.utility.get_transform_world_to_local_old(sphere_volume.name)
    (
        global_translation_new,
        global_rotation_new,
    ) = gate.geometry.utility.get_transform_world_to_local(sphere_volume)
    print("***********************************************************")
    print(f"World to local transform for volume {sphere_volume.name}: ")
    print("global_translation_old, global_rotation_old: ")
    print(global_translation_old, global_rotation_old)
    print("global_translation_new, global_rotation_new: ")
    print(global_translation_new, global_rotation_new)
    print("***********************************************************")


def check_mat(se):
    pathFile = pathlib.Path(__file__).parent.resolve()

    # explicit check overlap (already performed during initialize)
    print("check overlap with verbose")
    se.check_volumes_overlap(verbose=True)
    sim = se.simulation

    # print info material db
    dbn = sim.volume_manager.dump_material_database_names()
    mnist = se.volume_engine.get_database_material_names("NIST")
    mdb = se.volume_engine.get_database_material_names(
        pathFile.parent / "data" / "GateMaterials.db"
    )
    dm = se.volume_engine.dump_build_materials()
    print("Material info:")
    print("\t databases    :", dbn)
    print("\t mat in NIST  :", len(mnist), mnist)
    print("\t mat in db    :", mdb)
    print("\t defined mat  :", dm)

    print("dbn", dbn)
    assert dbn == [pathFile.parent / "data" / "GateMaterials.db"]
    # assert len(mnist) == 308  # Geant4 11.02
    assert len(mnist) == 309  # Geant4 11.1
    assert mdb == [
        "Vacuum",
        "Aluminium",
        "Uranium",
        "Silicon",
        "Germanium",
        "Yttrium",
        "Gadolinium",
        "Lutetium",
        "Tungsten",
        "Lead",
        "Bismuth",
        "NaI",
        "NaITl",
        "PWO",
        "BGO",
        "LSO",
        "Plexiglass",
        "GSO",
        "LuAP",
        "YAP",
        "Water",
        "Quartz",
        "Breast",
        "Air",
        "Glass",
        "Scinti-C9H10",
        "LuYAP-70",
        "LuYAP-80",
        "Plastic",
        "CZT",
        "Lung",
        "Polyethylene",
        "PVC",
        "SS304",
        "PTFE",
        "LYSO",
        "Body",
        "Muscle",
        "LungMoby",
        "SpineBone",
        "RibBone",
        "Adipose",
        "Blood",
        "Heart",
        "Kidney",
        "Liver",
        "Lymph",
        "Pancreas",
        "Intestine",
        "Skull",
        "Cartilage",
        "Brain",
        "Spleen",
        "Testis",
        "PMMA",
    ]
    assert dm == ["G4_AIR", "G4_WATER", "Lead", "Lung", "G4_LUCITE"]
    assert len(se.volume_engine.dump_build_materials()) == 5


if __name__ == "__main__":
    pathFile = pathlib.Path(__file__).parent.resolve()
    paths = utility.get_default_test_paths(__file__)

    # create the simulation
    sim = gate.Simulation()
    print(f"Volumes types: {sim.volume_manager.dump_volume_types()}")

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.store_json_archive = True
    sim.output_dir = paths.output / "test007"

    # add a material database
    sim.volume_manager.add_material_database(
        pathFile.parent / "data" / "GateMaterials.db"
    )

    #  change world size
    m = gate.g4_units.m
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]

    # add a simple volume
    waterbox = sim.add_volume("Box", "Waterbox")
    cm = gate.g4_units.cm
    waterbox.size = [60 * cm, 60 * cm, 60 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 35 * cm]
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]  # blue

    # another (child) volume with rotation
    mm = gate.g4_units.mm
    sheet = sim.add_volume("Box", "Sheet")
    sheet.size = [30 * cm, 30 * cm, 2 * mm]
    sheet.mother = "Waterbox"
    sheet.material = "Lead"
    r = Rotation.from_euler("x", 33, degrees=True)
    center = [0 * cm, 0 * cm, 10 * cm]
    t = gate.geometry.utility.get_translation_from_rotation_with_center(r, center)
    sheet.rotation = r.as_matrix()
    sheet.translation = t + [0 * cm, 0 * cm, -18 * cm]
    sheet.color = [1, 0, 0, 1]  # red

    # A sphere
    sph = sim.add_volume("Sphere", "mysphere")
    sph.rmax = 5 * cm
    sph.mother = "Waterbox"
    sph.translation = [0 * cm, 0 * cm, -8 * cm]
    sph.material = "Lung"
    sph.color = [0.5, 1, 0.5, 1]  # kind of green
    sph.toto = "nothing"  # ignored, should raise a warning

    # A ...thing ?
    trap = sim.add_volume("Trap", "mytrap")
    trap.mother = "Waterbox"
    trap.translation = [0, 0, 15 * cm]
    trap.material = "G4_LUCITE"

    # default source for tests
    source = sim.add_source("GenericSource", "Default")
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    source.particle = "proton"
    source.energy.mono = 240 * MeV
    source.position.radius = 1 * cm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 2500 * Bq

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    # run timing
    sec = gate.g4_units.second
    sim.run_timing_intervals = [
        [0, 0.5 * sec]
        # ,[0.5 * sec, 1.2 * sec]
    ]

    print(sim)

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")
    # sim.g4_com("/run/verbose 2")
    # sim.g4_com("/event/verbose 2")
    # sim.g4_com("/tracking/verbose 1")

    # start simulation
    sim.user_hook_after_init = check_mat
    sim.user_hook_after_run = user_hook_volume
    # should be sim.run(start_new_process=True)
    # but for testing, we currently set it to False
    sim.run(start_new_process=False)

    # print results at the end
    print(stats)

    # check
    stats_ref = gate.actors.miscactors.SimulationStatisticsActor(name="ref")
    c = stats_ref.counts
    c.runs = 1
    c.events = 1280
    c.tracks = 17034  # 25668
    c.steps = 78096  # 99465
    # stats_ref.pps = 506.6
    sec = gate.g4_units.second
    c.duration = 2.5267 * sec
    print("-" * 80)
    is_ok = utility.assert_stats(stats, stats_ref, 0.15)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import uproot
from opengate.tests import utility
import opengate as gate
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt, TreatmentPlanSource
from scipy.spatial.transform import Rotation
import itk
import matplotlib.pyplot as plt
import os
import numpy as np


def calculate_mean_unc(edep_arr, unc_arr, edep_thresh_rel=0.7):
    edep_max = np.amax(edep_arr)
    mask = edep_arr > edep_max * edep_thresh_rel
    unc_used = unc_arr[mask]
    unc_mean = np.mean(unc_used)

    return unc_mean


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test044_pbs", output_folder="test059"
    )
    output_path = paths.output
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 123654789
    sim.random_engine = "MersenneTwister"
    sim.number_of_threads = 1
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    mrad = gate.g4_units.mrad
    s = gate.g4_units.s

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    ## FIRST DETECTOR ##
    # box
    # translation and rotation like in the Gate macro
    box1 = sim.add_volume("Box", "box1")
    box1.size = [200 * mm, 200 * mm, 1050 * mm]
    # box1.translation = [0 * cm, 0 * cm, 52.5 * cm]
    # box1.rotation = Rotation.from_euler('y',-90,degrees=True).as_matrix()
    box1.translation = [0 * cm, 0 * cm, 52.5 * cm]
    box1.material = "Vacuum"
    box1.color = [0, 0, 1, 1]

    # phantoms
    m = Rotation.identity().as_matrix()

    phantom1_s1 = sim.add_volume("Box", "phantom1_s1")
    phantom1_s1.mother = "box1"
    phantom1_s1.size = [100 * mm, 100 * mm, 50 * mm]
    phantom1_s1.translation = [-50 * mm, 0 * mm, -500 * mm]
    phantom1_s1.rotation = m
    phantom1_s1.material = "G4_WATER"
    phantom1_s1.color = [1, 0, 1, 1]

    phantom1_s2 = sim.add_volume("Box", "phantom1_s2")
    phantom1_s2.mother = "box1"
    phantom1_s2.size = [100 * mm, 100 * mm, 50 * mm]
    phantom1_s2.translation = [50 * mm, 0 * mm, -500 * mm]
    phantom1_s2.rotation = m
    phantom1_s2.material = "G4_WATER"
    phantom1_s2.color = [1, 0, 1, 1]

    # add dose actor
    dose1_s1 = sim.add_actor("DoseActor", "edep_11")
    filename = "phantom1_s1.mhd"
    dose1_s1.output_filename = filename
    dose1_s1.attached_to = "phantom1_s1"
    dose1_s1.size = [25, 25, 1]
    dose1_s1.spacing = [4.0, 4.0, 50.0]
    dose1_s1.hit_type = "random"
    dose1_s1.edep_uncertainty.active = True
    # dose1_s1.ste_of_mean = True

    # add dose actor
    dose1_s2 = sim.add_actor("DoseActor", "edep_12")
    filename = "phantom1_s2.mhd"
    dose1_s2.output_filename = filename
    dose1_s2.attached_to = "phantom1_s2"
    dose1_s2.size = [25, 25, 1]
    dose1_s2.spacing = [4.0, 4.0, 50.0]
    dose1_s2.hit_type = "random"
    dose1_s2.edep_uncertainty.active = True
    # # PhaseSpace Actor
    # Phsp_act = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    # Phsp_act.attached_to = phantom1_s2.name
    # Phsp_act.attributes = [
    #     "KineticEnergy",
    #     "EventID",
    #     "ThreadID",
    # ]
    # Phsp_act.output = output_path / "test058_MT.root"
    # Phsp_act.debug = False

    ## SECOND DETECTOR ##
    # box
    # translation and rotation like in the Gate macro
    box2 = sim.add_volume("Box", "box2")
    box2.size = [200 * mm, 200 * mm, 1050 * mm]
    # box2.rotation = Rotation.from_euler('y',90,degrees=True).as_matrix()
    box2.translation = [40 * cm, 0 * cm, 52.5 * cm]
    # box2.translation = [-52.5* cm, 0 * cm, 40 * cm]
    box2.material = "Vacuum"
    box2.color = [0, 0, 1, 1]

    # phantoms
    m = Rotation.identity().as_matrix()

    phantom2_s1 = sim.add_volume("Box", "phantom2_s1")
    phantom2_s1.mother = "box2"
    phantom2_s1.size = [100 * mm, 100 * mm, 50 * mm]
    phantom2_s1.translation = [-50 * mm, 0 * mm, -500 * mm]
    phantom2_s1.rotation = m
    phantom2_s1.material = "G4_WATER"
    phantom2_s1.color = [1, 0, 1, 1]

    phantom2_s2 = sim.add_volume("Box", "phantom2_s2")
    phantom2_s2.mother = "box2"
    phantom2_s2.size = [100 * mm, 100 * mm, 50 * mm]
    phantom2_s2.translation = [50 * mm, 0 * mm, -500 * mm]
    phantom2_s2.rotation = m
    phantom2_s2.material = "G4_WATER"
    phantom2_s2.color = [1, 0, 1, 1]

    # add dose actor
    dose2_s1 = sim.add_actor("DoseActor", "edep_21")
    filename = "phantom2_s1.mhd"
    dose2_s1.output_filename = filename
    dose2_s1.attached_to = "phantom2_s1"
    dose2_s1.size = [25, 25, 1]
    dose2_s1.spacing = [4.0, 4.0, 50.0]
    dose2_s1.hit_type = "random"
    dose2_s1.edep_uncertainty.active = True

    dose2_s2 = sim.add_actor("DoseActor", "edep_22")
    filename = "phantom2_s2.mhd"
    dose2_s2.output_filename = filename
    dose2_s2.attached_to = "phantom2_s2"
    dose2_s2.size = [25, 25, 1]
    dose2_s2.spacing = [4.0, 4.0, 50.0]
    dose2_s2.hit_type = "random"
    dose2_s2.edep_uncertainty.active = True

    ## TPS SOURCE ##
    # beamline model
    beamline = BeamlineModel()
    beamline.name = None
    beamline.radiation_types = "proton"

    # polinomial coefficients
    beamline.energy_mean_coeffs = [1, 0]
    beamline.energy_spread_coeffs = [0.4417036946562556]
    beamline.sigma_x_coeffs = [2.3335754]
    beamline.theta_x_coeffs = [2.3335754e-3]
    beamline.epsilon_x_coeffs = [0.00078728e-3]
    beamline.sigma_y_coeffs = [1.96433431]
    beamline.theta_y_coeffs = [0.00079118e-3]
    beamline.epsilon_y_coeffs = [0.00249161e-3]

    # tps
    nSim = 40000  # particles to simulate per beam
    print("--- Flat spots distribution ---")
    tps_flat = sim.add_source("TreatmentPlanPBSource", "flat")
    tps_flat.n = nSim
    tps_flat.end_time = 0.5 * s
    tps_flat.flat_generation = True
    tps_flat.beam_model = beamline
    tps_flat.plan_path = ref_path / "TreatmentPlan2Spots_flat_gen_test.txt"
    tps_flat.beam_nr = 1
    tps_flat.gantry_rot_axis = "x"
    tps_flat.particle = "proton"
    tps_flat.position.translation = [0 * cm, 0 * cm, -30 * cm]

    print("--- Proportional spots distribution ---")
    tps_non_flat = sim.add_source("TreatmentPlanPBSource", "proportional")
    tps_non_flat.n = nSim
    tps_non_flat.start_time = 0.5 * s
    tps_non_flat.flat_generation = False
    tps_non_flat.beam_model = beamline
    tps_non_flat.plan_path = ref_path / "TreatmentPlan2Spots_flat_gen_test.txt"
    tps_non_flat.beam_nr = 1
    tps_non_flat.gantry_rot_axis = "x"
    tps_non_flat.particle = "proton"
    tps_non_flat.position.translation = [40 * cm, 0 * cm, 0 * cm]

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.set_production_cut("world", "all", 1000 * km)
    # sim.set_user_limits("phantom_a_2","max_step_size",1,['proton'])

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # start simulation
    sim.run()

    # print results at the end
    print(stats)

    # ----------------------------------------------------------------------------------------------------------------
    # tests
    # with and without flat generation, the result in terms of dose should be identical
    d_names = ["edep_11", "edep_12", "edep_21", "edep_22"]
    dose_actors = [sim.get_actor(d) for d in d_names]
    test = True

    # check that the dose output is the same
    print("--- Dose image spot 0 ---")
    test = (
        utility.assert_images(
            output_path / dose_actors[0].get_output_path("edep"),
            output_path / dose_actors[2].get_output_path("edep"),
            stats,
            tolerance=70,
            ignore_value_data2=0,
        )
        and test
    )

    print("--- Dose image spot 1 ---")
    test = (
        utility.assert_images(
            output_path / dose_actors[1].get_output_path("edep"),
            output_path / dose_actors[3].get_output_path("edep"),
            stats,
            tolerance=70,
            ignore_value_data2=0,
            sum_tolerance=5.2,
        )
        and test
    )

    # check that output with flat distribution has better statistics for the spot with less particles
    unc_vec = []
    for d, name in zip(dose_actors, d_names):
        edep_img = itk.imread(paths.output / d.get_output_path("edep"))
        edep_arr = itk.GetArrayViewFromImage(edep_img)
        unc_img = itk.imread(paths.output / d.get_output_path("edep_uncertainty"))
        unc_arr = itk.GetArrayFromImage(unc_img)
        # fig = utility.plot2D(unc_arr[0,:,:], d_actor, show=True)
        mean_unc = calculate_mean_unc(edep_arr, unc_arr, edep_thresh_rel=0.2)
        unc_vec.append(mean_unc)
        print(f"mean uncertainty {name}: {mean_unc}")

    test = unc_vec[1] < unc_vec[3] and test

    utility.test_ok(test)



from collections import deque
from itertools import chain

import awkward as ak
import numpy as np
import pandas as pd
import os
import time
import logging
import uproot
import sys

logger = logging.getLogger(__name__)


class ChunkSizeTooSmallError(Exception):
    pass


class CoincidenceOutputFile:
    def __init__(self, file_path, file_format):
        assert file_format in ["root", "hdf5"]
        self.file_path = file_path
        self.format = file_format
        self.empty = True

        if self.format == "root":
            # recreate() deletes the file that may already exist with the same name.
            self.file = uproot.recreate(self.file_path)
        elif self.format == "hdf5":
            if os.path.exists(self.file_path):
                os.remove(self.file_path)

    def add(self, coincidences):
        if len(coincidences) == 0:
            return
        # dtype "object" (used for strings) is not accepted by extend() function in uproot,
        # need to convert to categorical.
        for col in coincidences.columns:
            if coincidences[col].dtype == "object":
                coincidences[col] = pd.Categorical(coincidences[col])
        table_name = "Coincidences"
        if self.format == "root":
            if self.empty:
                self.file[table_name] = coincidences
            else:
                self.file[table_name].extend(coincidences)
        elif self.format == "hdf5":
            coincidences.to_hdf(
                self.file_path,
                key=table_name,
                mode="a",
                format="table",
                append=True,
                index=False,
            )
        self.empty = False

    def close(self):
        if self.format == "root":
            self.file.close()


def coincidences_sorter(
    singles_tree,
    time_window,
    policy,
    min_transaxial_distance,
    transaxial_plane,
    max_axial_distance,
    chunk_size=100000,
    return_type="dict",
    output_file_path=None,
    output_file_format="root",
):
    """
    Sort singles and detect coincidences.
    :param singles_tree: input tree of singles (root format)
    :param time_window: time windows in G4 units (ns)
    :param policy: coincidence detection policy, one of:
            "removeMultiples"
            "takeAllGoods"
            "takeWinnerOfGoods"
            "takeIfOnlyOneGood"
            "takeWinnerIfIsGood"
            "takeWinnerIfAllAreGoods"
    :param min_transaxial_distance: minimum transaxial distance between the two singles of a coincidence
    :param transaxial_plane: "xy", "yz", or "xz"
    :param max_axial_distance: maximum axial distance between the two singles of a coincidence
    :param chunk_size: singles are processed by this chunk size
    :param return_type: "dict" or "pd"
    :param output_file_path: if provided, the coincidences will be saved to the given file path
    :param output_file_format: "root" or "hdf5"
    :return: if output_file_path is given, the return value is None, otherwise the coincidences are returned
             as a dict of events (return_type "dict") or a pandas DataFrame (return_type "pd")

    Chunk size is important for very large root file to avoid loading everything in memory at once

    DEV NOTES:
    1) TODO: add option allDigiOpenCoincGate=false, so far only for allDigiOpenCoincGate=true
    """

    # Check availability of the necessary branches in the root file
    required_branches = {
        "EventID",
        "GlobalTime",
        "PreStepUniqueVolumeID",
        "TotalEnergyDeposit",
        "PostPosition_X",
        "PostPosition_Y",
        "PostPosition_Z",
    }
    missing_branches = required_branches - set(singles_tree.keys())
    if missing_branches:
        if len(missing_branches) == 1:
            raise ValueError(
                f"Required branch {missing_branches.pop()} is missing in singles tree"
            )
        else:
            raise ValueError(
                f"Required branches {missing_branches} are missing in singles tree"
            )

    # Check validity of policy parameter
    policy_functions = {
        "removeMultiples": remove_multiples,
        "takeAllGoods": take_all_goods,
        "takeWinnerOfGoods": take_winner_of_goods,
        "takeIfOnlyOneGood": take_if_only_one_good,
        "takeWinnerIfIsGood": take_winner_if_is_good,
        "takeWinnerIfAllAreGoods": take_winner_if_all_are_goods,
    }
    if policy not in policy_functions:
        raise ValueError(
            f"Unknown policy '{policy}', must be one of {policy_functions.keys()}"
        )

    # Check validity of return_type or output_file_format and output_file_path.
    if output_file_path is None:
        known_return_types = ["dict", "pd"]
        if return_type not in known_return_types:
            raise ValueError(
                f"Unknown return type '{return_type}', must be one of {known_return_types}"
            )
    else:
        known_output_formats = ["root", "hdf5"]
        if output_file_format not in known_output_formats:
            raise ValueError(
                f"Unknown output file format '{output_file_format}', must be one of {known_output_formats}"
            )
        # Saving to a HDF5 file in append mode requires pytables>=3.10 if numpy>=2.0 is used,
        # but that version of pytables is only supported from Python 3.10 onwards
        if output_file_format == "hdf5" and sys.version_info[1] < 10:
            raise NotImplementedError(
                "HDF5 output is only supported in Python 3.10 or newer"
            )
        if not output_file_path:
            raise ValueError(f"Output file path has not been provided")

    # Since singles in the root file are not guaranteed to be sorted by GlobalTime
    # (especially in case of multithreaded simulation), singles in one chunk
    # may be more recent than the singles in the next chunk.
    # If that's the case, the chunk size must be increased for successful coincidence sorting.
    original_chunk_size = chunk_size
    max_num_chunk_size_increases = 10
    num_chunk_size_increases = 0
    processing_finished = False
    start = time.time()
    time_lost = 0
    while (
        not processing_finished
        and num_chunk_size_increases < max_num_chunk_size_increases
    ):
        try:
            if output_file_path:
                output_file = CoincidenceOutputFile(
                    output_file_path, output_file_format
                )

            # A double-ended queue is used as a FIFO to store the current and the next chunk of singles
            queue = deque()
            num_singles = 0
            coincidences_to_transfer = None
            coincidences_to_return = []
            for chunk in singles_tree.iterate(step_size=chunk_size):
                num_singles_in_chunk = len(chunk)
                # Convert chunk to pandas DataFrame
                chunk_pd = ak.to_dataframe(chunk)
                # Add a temporary column to assist in applying the various policies
                chunk_pd["SingleIndex"] = range(
                    num_singles, num_singles + num_singles_in_chunk
                )
                # Add chunk to the right of the queue
                queue.append(chunk_pd)
                # Process a chunk, unless only one has been read so far
                if len(queue) > 1:
                    coincidences = process_chunk(queue, time_window)
                    # Before filtering coincidences, we want to make sure that we have all
                    # coincidences that belong to the same time window (same SingleIndex1 value).
                    # When processing the next chunk of singles, we may still find one or more
                    # coincidences that belong to the same time window as the last coincidence so far.
                    #
                    # All coincidences that have a SingleIndex1 value different from the last one
                    # can already be filtered, the others will be transferred to be filtered in
                    # the next iteration.
                    coincidences_to_filter = coincidences.loc[
                        coincidences["SingleIndex1"]
                        != coincidences["SingleIndex1"].iloc[-1]
                    ].reset_index(drop=True)
                    if coincidences_to_transfer is not None:
                        coincidences_to_filter = pd.concat(
                            [coincidences_to_transfer, coincidences_to_filter],
                            axis=0,
                            ignore_index=True,
                        )
                    coincidences_to_transfer = coincidences.loc[
                        coincidences["SingleIndex1"]
                        == coincidences["SingleIndex1"].iloc[-1]
                    ].reset_index(drop=True)
                    # Apply policy
                    filtered_coincidences = policy_functions[policy](
                        coincidences_to_filter,
                        min_transaxial_distance,
                        transaxial_plane,
                        max_axial_distance,
                    )
                    # Remove the temporary SingleIndex columns
                    filtered_coincidences = filtered_coincidences.drop(
                        columns=["SingleIndex1", "SingleIndex2"]
                    )
                    if output_file_path:
                        output_file.add(filtered_coincidences)
                    else:
                        coincidences_to_return.append(filtered_coincidences)
                    # Remove processed chunk from the left of the queue
                    queue.popleft()
                num_singles += num_singles_in_chunk

            # At this point, all chunks have been read. Now process the last chunk.
            coincidences_to_filter = process_chunk(queue, time_window)
            if coincidences_to_transfer is not None:
                coincidences_to_filter = pd.concat(
                    [coincidences_to_transfer, coincidences_to_filter],
                    axis=0,
                    ignore_index=True,
                )
            filtered_coincidences = policy_functions[policy](
                coincidences_to_filter,
                min_transaxial_distance,
                transaxial_plane,
                max_axial_distance,
            )
            # Remove the temporary SingleIndex columns
            filtered_coincidences = filtered_coincidences.drop(
                columns=["SingleIndex1", "SingleIndex2"]
            )
            if output_file_path:
                output_file.add(filtered_coincidences)
            else:
                coincidences_to_return.append(filtered_coincidences)

            processing_finished = True

        except ChunkSizeTooSmallError:
            # Double chunk size and start all over again
            chunk_size *= 2
            num_chunk_size_increases += 1
            time_lost = time.time() - start

        finally:
            if output_file_path:
                output_file.close()

    if not processing_finished:
        # Coincidence sorting has failed, even after repeated increases of the chunk size
        raise ChunkSizeTooSmallError

    if num_chunk_size_increases > 0:
        time_reduction = time_lost / (time.time() - start)
        if time_reduction >= 0.01:
            logger.warning(
                f"Coincidence sorting execution time can be reduced by {time_reduction*100:.02f}% by using chunk_size {chunk_size} instead of {original_chunk_size}"
            )

    if output_file_path is None:
        # Combine all coincidences from all chunks into a single pandas DataFrame
        coincidences_to_return = pd.concat(
            coincidences_to_return, axis=0, ignore_index=True
        )
        if return_type == "dict":
            return coincidences_to_return.to_dict(orient="list")
        elif return_type == "pd":
            return coincidences_to_return


def process_chunk(queue, time_window):
    """
    Processes singles in the chunk queue[0],
    possibly transferring some of those singles to the next chunk queue[1].
    """
    chunk = queue[0]
    next_chunk = queue[1] if len(queue) > 1 else None

    t1 = chunk["GlobalTime"]
    t1_min = np.min(t1)
    t1_max = np.max(t1)

    if next_chunk is not None:
        t2 = next_chunk["GlobalTime"]
        t2_min = np.min(t2)
        t2_max = np.max(t2)
        # Require that the next chunk's time interval is later than
        # the current chunk's time interval (overlap (t2_min < t1_max) is allowed).
        if not (t2_min > t1_min and t2_max > t1_max):
            raise ChunkSizeTooSmallError

    # Find coincidences in the current chunk
    coincidences = run_coincidence_detection_in_chunk(chunk, time_window)

    if next_chunk is not None:
        # If there are singles in the current chunk that are beyond t_min
        # of the next chunk minus time_window, then we want to process them
        # in the next chunk as well.

        # Remove any coincidences for which those singles have opened the time window
        # in the current chunk.
        coincidences = coincidences.loc[
            coincidences["GlobalTime1"] < t2_min - time_window
        ].reset_index(drop=True)
        # Find and add singles to be treated in the next chunk
        singles_to_transfer = chunk.loc[
            chunk["GlobalTime"] >= t2_min - time_window
        ].reset_index(drop=True)
        queue[1] = pd.concat([singles_to_transfer, next_chunk], axis=0)

    return coincidences


def run_coincidence_detection_in_chunk(chunk, time_window):
    """
    Detects coincidences between singles in the given chunk, excluding coincidences
    between singles in the same volume.
    """
    # Add a temporary column containing hash values of the strings identifying the volumes,
    # to assist in excluding coincidences between singles in the same volume
    # (calculating and comparing hash values is much faster than comparing strings).
    chunk["VolumeIDHash"] = pd.util.hash_pandas_object(
        chunk["PreStepUniqueVolumeID"], index=False
    )
    time_np = chunk["GlobalTime"].to_numpy()
    # Sort the time values chronologically (singles in the chunk may not be in chronological order).
    time_np_sorted_indices = np.argsort(time_np)
    time_np = time_np[time_np_sorted_indices]
    # Create NumPy arrays for the indices of the first and second single in each coincidence.
    indices1 = np.zeros((0,), dtype=np.int32)
    indices2 = np.zeros((0,), dtype=np.int32)
    # Repeatedly consider single pairs that are 1, 2, etc. positions apart in the sorted list,
    # until no more coincidences can be found.
    offset = 1
    while True:
        delta_time = time_np[offset:] - time_np[:-offset]
        indices = np.nonzero(delta_time <= time_window)[0]
        if len(indices) == 0:
            break
        indices1 = np.concatenate((indices1, indices))
        indices2 = np.concatenate((indices2, indices + offset))
        offset += 1
    # Combine indices1 and indices2 in a two-column NumPy array.
    indices12 = np.vstack((indices1, indices2)).T
    # Sort rows such that index1 is increasing, and for rows with the same
    # index1 value, index2 is increasing.
    indices12 = indices12[np.lexsort((indices12[:, 1], indices12[:, 0]))]
    # Create dictionaries to create the names of the columns in the coincidences pandas DataFrame.
    rename_dict1 = {name: name + "1" for name in chunk.columns.tolist()}
    rename_dict2 = {name: name + "2" for name in chunk.columns.tolist()}
    # Create pandas DataFrames with the sorted singles that are part of the coincidences.
    coincidence_singles1 = (
        chunk.iloc[time_np_sorted_indices[indices12[:, 0]]]
        .rename(columns=rename_dict1)
        .reset_index(drop=True)
    )
    coincidence_singles2 = (
        chunk.iloc[time_np_sorted_indices[indices12[:, 1]]]
        .rename(columns=rename_dict2)
        .reset_index(drop=True)
    )
    # Combine both DataFrames into one, interleaving the columns.
    interleaved_columns = list(
        chain(*zip(coincidence_singles1.columns, coincidence_singles2.columns))
    )
    coincidences = pd.concat([coincidence_singles1, coincidence_singles2], axis=1)[
        interleaved_columns
    ]
    # Remove coincidences between singles in the same volume.
    coincidences = coincidences.loc[
        coincidences["VolumeIDHash1"] != coincidences["VolumeIDHash2"]
    ].reset_index(drop=True)
    # Remove the temporary volume ID hash columns.
    coincidences = coincidences.drop(columns=["VolumeIDHash1", "VolumeIDHash2"])

    return coincidences


def remove_multiples(
    coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
):
    # Remove multiple coincidences.
    # Remaining coincidences are the ones that are alone in their time window.
    coincidences_multiples_removed = filter_multi(coincidences)
    # Only keep coincidences that comply with the minimum transaxial distance
    # and the maximum axial distance.
    coincidences_output = filter_goods(
        coincidences_multiples_removed,
        min_transaxial_distance,
        transaxial_plane,
        max_axial_distance,
    )
    return coincidences_output


def take_all_goods(
    coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
):
    # Only keep coincidences that comply with the minimum transaxial distance
    # and the maximum axial distance.
    return filter_goods(
        coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
    )


def take_winner_of_goods(
    coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
):
    # Of all the "good" coincidences in a window, only keep the one with highest energy.

    coincidences_goods = filter_goods(
        coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
    )
    return filter_max_energy(coincidences_goods)


def take_if_only_one_good(
    coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
):
    # Keep only the "good" coincidences, then remove multiples from the remaining coincidences.
    # As a result, the only coincidences remainining are the ones that are the only "good" one
    # in their coincidence window.

    coincidences_goods = filter_goods(
        coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
    )

    return filter_multi(coincidences_goods)


def take_winner_if_is_good(
    coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
):
    # Only keep the coincidence with max energy in each time window, then remove
    # coincidences that are not "good".
    coincidences_winner = filter_max_energy(coincidences)
    return filter_goods(
        coincidences_winner,
        min_transaxial_distance,
        transaxial_plane,
        max_axial_distance,
    )


def take_winner_if_all_are_goods(
    coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
):
    filtered_coincidences = filter_goods(
        coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
    )

    # SingleIndex1 is the index of the single that has opened the time window for a coincidence.
    # Coincidences with the same value of SingleIndex1 belong to the same time window.

    # Count the number of occurrences of all SingleIndex1 values before and after filtering the goods.
    counts = coincidences["SingleIndex1"].value_counts()
    filtered_counts = filtered_coincidences["SingleIndex1"].value_counts()

    # Find the SingleIndex1 values that appear in both count lists.
    common_index1_values = counts.index.intersection(filtered_counts.index)
    # Find the SingleIndex1 values that have the same count value before and after filtering the goods.
    index1_values_with_matching_counts = common_index1_values[
        counts[common_index1_values] == filtered_counts[common_index1_values]
    ]
    # Only keep the coincidences if the count in their time window has not changed
    # by filtering the goods ("all are good").
    filtered_coincidences = filtered_coincidences[
        filtered_coincidences["SingleIndex1"].isin(index1_values_with_matching_counts)
    ]

    # In each time window, take the "winner" with maximal sum of singles deposited energy.
    return filter_max_energy(filtered_coincidences)


def filter_multi(coincidences):

    # Coincidences with the same value of SingleIndex1 belong to the same time window.
    # Remove multiples by removing coincidences where the first single appears more than once.
    return coincidences[~coincidences["SingleIndex1"].duplicated(keep=False)]


def filter_goods(
    coincidences, min_transaxial_distance, transaxial_plane, max_axial_distance
):
    # Calculate the transaxial distance between the singles in each coincidence.
    td = transaxial_distance(coincidences, transaxial_plane)
    # Remove coincidences of which the transaxial distance is below threshold.
    filtered_coincidences = coincidences.loc[td >= min_transaxial_distance].reset_index(
        drop=True
    )
    # Calculate the axial distance between the singles in each remaining coincidence.
    ad = axial_distance(filtered_coincidences, transaxial_plane)
    # Remove coincidences of which the axial distance is above the threshold.
    return filtered_coincidences.loc[ad <= max_axial_distance].reset_index(drop=True)


def filter_max_energy(coincidences):

    filtered_coincidences = coincidences.copy(deep=True)
    # Add a temporary column containing the sum of deposited energy of both singles.
    filtered_coincidences["TotalEnergyInCoincidence"] = (
        filtered_coincidences["TotalEnergyDeposit1"]
        + filtered_coincidences["TotalEnergyDeposit2"]
    )
    # Of all the coincidences with the same "SingleIndex1" value (belonging to the same time window),
    # only keep the one with the highest value in column "TotalEnergyInCoincidence".
    filtered_coincidences = filtered_coincidences.loc[
        filtered_coincidences.groupby("SingleIndex1")[
            "TotalEnergyInCoincidence"
        ].idxmax()
    ]
    # Remove the temporary column.
    return filtered_coincidences.drop(columns=["TotalEnergyInCoincidence"])


def transaxial_distance(coincidences, transaxial_plane):

    if transaxial_plane not in ("xy", "yz", "xz"):
        raise ValueError(
            f"Invalid transaxial_plane: '{transaxial_plane}'. Expected one of 'xy', 'yz' or 'xz'."
        )
    # Extract the first transaxial plane coordinate values for both singles from all coincidences.
    a1, a2 = [
        coincidences[f"PostPosition_{transaxial_plane[0].upper()}{n}"].to_numpy()
        for n in (1, 2)
    ]
    # Extract the second transaxial plane coordinate values for both singles from all coincidences.
    b1, b2 = [
        coincidences[f"PostPosition_{transaxial_plane[1].upper()}{n}"].to_numpy()
        for n in (1, 2)
    ]

    return np.sqrt((a1 - a2) ** 2 + (b1 - b2) ** 2)


def axial_distance(coincidences, transaxial_plane):

    if transaxial_plane not in ("xy", "yz", "xz"):
        raise ValueError(
            f"Invalid transaxial_plane: '{transaxial_plane}'. Expected one of 'xy', 'yz' or 'xz'."
        )
    # Determine the coordinate perpendicular to the transaxial plane.
    axial_coordinate = (set("xyz") - set(transaxial_plane)).pop().upper()
    # Extract the coordinate values for both singles from all coincidences.
    a1, a2 = [
        coincidences[f"PostPosition_{axial_coordinate}{n}"].to_numpy() for n in (1, 2)
    ]

    return np.abs(a1 - a2)


def copy_tree_for_dump(input_tree):
    branches = {}
    for key in input_tree.keys():
        branches[key] = input_tree.arrays([key], library="np")[key]
    return branches



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests.src import test022_half_life


def main():
    test022_half_life.test022_half_life(n_threads=3)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test090_helpers import *
from opengate import g4_units

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test090_spect_config_6"
    )

    data_path = paths.data

    # units
    deg = g4_units.deg
    Bq = g4_units.Bq
    cm = g4_units.cm
    mm = g4_units.mm
    sec = g4_units.s

    sc = SPECTConfig()
    sc.simu_name = "test090"
    sc.output_folder = paths.output
    sc.number_of_threads = 1
    # detectors
    sc.detector_config.model = "intevo"
    sc.detector_config.collimator = "melp"
    sc.detector_config.number_of_heads = 2
    sc.detector_config.digitizer_channels = intevo.get_default_energy_windows("lu177")
    # phantom
    sc.phantom_config.image = data_path / "iec_5mm.mhd"
    sc.phantom_config.labels = data_path / "iec_5mm_labels.json"
    sc.phantom_config.material_db = data_path / "iec_5mm.db"
    # source
    sc.source_config.image = data_path / "iec_5mm_activity.mhd"
    sc.source_config.radionuclide = "177lu"
    sc.source_config.total_activity = 5e4 * Bq
    # acquisition
    sc.acquisition_config.radius = 300 * mm
    sc.acquisition_config.duration = 30 * sec
    sc.acquisition_config.number_of_angles = 3
    # ff
    sc.free_flight_config.primary_activity = 1e5 * Bq
    sc.free_flight_config.max_rejection = 10000
    sc.free_flight_config.angle_tolerance = 15 * deg
    sc.free_flight_config.forced_direction_flag = False

    # create the simulation
    print(sc)
    sim = gate.Simulation()
    sc.setup_simulation(sim, visu=False)

    # write in a JSON file
    sc.to_json(paths.output / "spect_config.json")

    # read into another sc
    print()
    sc2 = SPECTConfig().from_json(paths.output / "spect_config.json")
    print(sc2)

    print()
    b = sc == sc2
    utility.print_test(b, "Test equality between write and read JSON files")
    is_ok = b

    sim = gate.Simulation()
    sc2.setup_simulation(sim, visu=False)
    b = sc == sc2
    utility.print_test(
        b, "Test equality between write and read JSON files (after setup)"
    )
    is_ok = b and is_ok

    utility.test_ok(is_ok)



import uproot
import numpy as np
import numbers
from scipy.spatial.transform import Rotation
from box import Box
import sys

import opengate_core as g4
from ..exception import fatal, warning
from .generic import SourceBase
from ..base import process_cls


class PhaseSpaceSourceGenerator:
    """
    Class that read phase space root file and extract position/direction/energy/weights of particles.
    Particles information will be copied to the c++ side to be used as a source
    """

    def __init__(self, tid):
        self.phsp_source = None
        self.tid = tid
        self.root_file = None
        self.num_entries = 0
        self.cycle_count = 0
        self.cycle_changed_flag = False
        # used during generation
        self.batch = None
        self.points = None
        self.current_index = 0
        self.w = None

    def initialize(self, phsp_source):
        self.phsp_source = phsp_source
        self.name = phsp_source.name
        # set the keys and entry start
        self.read_phsp_and_keys()

    def read_phsp_and_keys(self):
        # convert str like 1e5 to int
        self.phsp_source.batch_size = int(self.phsp_source.batch_size)
        if self.phsp_source.batch_size < 1:
            fatal("PhaseSpaceSourceGenerator: Batch size should be > 0")

        if g4.IsMultithreadedApplication() and g4.G4GetThreadId() == -1:
            # do nothing for master thread
            return

        # open root file and get the first branch
        # FIXME could have an option to select the branch
        self.root_file = uproot.open(self.phsp_source.phsp_file)
        branches = self.root_file.keys()
        if len(branches) > 0:
            self.root_file = self.root_file[branches[0]]
        else:
            fatal(
                f"PhaseSpaceSourceGenerator: No usable branches in the root file {self.phsp_source.phsp_file}. Aborting."
            )
            sys.exit()

        self.num_entries = int(self.root_file.num_entries)

        # initialize the index to start
        tid = g4.G4GetThreadId()
        self.current_index = self.get_entry_start(self.phsp_source.entry_start)

        # initialize counters
        self.cycle_count = 0

    def get_entry_start(self, entry_start):
        if not g4.IsMultithreadedApplication():
            if not isinstance(entry_start, numbers.Number):
                fatal("entry_start must be a simple number is mono-thread mode")
            n = int(entry_start % self.num_entries)
            if entry_start > self.num_entries:
                warning(
                    f"In source {self.name} "
                    f"entry_start = {entry_start} while "
                    f"the phsp contains {self.num_entries}. "
                    f"We consider {n} instead (modulo)"
                )
            return n
        tid = g4.G4GetThreadId()

        if tid < 0:
            # no entry start needed for master thread
            return 0
        n_threads = g4.GetNumberOfRunningWorkerThreads()
        if isinstance(entry_start, numbers.Number):
            fatal(f"entry_start must be a list in multi-thread mode")
        if len(entry_start) != n_threads:
            fatal(
                f"Error: entry_start must be a vector of length the nb of threads, "
                f"but it is {len(entry_start)} instead of {n_threads}"
            )
        n = int(entry_start[self.tid] % self.num_entries)
        if entry_start[tid] > self.num_entries:
            warning(
                f"In source {self.name} "
                f"entry_start = {entry_start} (thread {tid}) "
                f"while the phsp contains {self.num_entries}. "
                f"We consider {n} instead (modulo)"
            )
        return n

    def generate(self, source, pid):
        """
        Main function that will be called from the cpp side every time a batch
        of particles should be created.
        Once created here, the particles are copied to cpp.
        (Yes maybe the copy could be avoided, but I did not manage to do it)
        """

        # warn phsp is recycled
        if self.cycle_changed_flag:
            warning(
                f"End of the phase-space {self.num_entries} elements, "
                f"restart from beginning. Cycle count = {self.cycle_count}"
            )
            self.cycle_changed_flag = False
            self.current_index = 0

        # read data from root tree
        current_batch_size = source.batch_size
        if self.current_index + source.batch_size >= self.num_entries:
            current_batch_size = self.num_entries - self.current_index
            self.cycle_count += 1
            self.cycle_changed_flag = True

        if source.verbose_batch:
            print(
                f"Thread {g4.G4GetThreadId()} "
                f"generate {current_batch_size} starting {self.current_index} "
                f" (phsp as n = {self.num_entries} entries)"
            )

        # read a batch of particles in the phsp (keep the reference)
        self.batch = self.root_file.arrays(
            entry_start=self.current_index,
            entry_stop=self.current_index + current_batch_size,
            library="numpy",
        )
        batch = self.batch
        self.current_index += current_batch_size

        # ensure encoding is float32
        for key in batch:
            # Convert to float32 if the array contains floating-point values
            if np.issubdtype(batch[key].dtype, np.floating):
                batch[key] = batch[key].astype(np.float32)
            else:
                if np.issubdtype(batch[key].dtype, np.integer):
                    batch[key] = batch[key].astype(np.int32)

        # set particle type
        if source.particle == "" or source.particle is None:
            # check if the keys for PDGCode are in the root file
            if source.PDGCode_key not in batch:
                fatal(
                    f"PhaseSpaceSource: no PDGCode key ({source.PDGCode_key}) "
                    f"in the phsp file and no source.particle"
                )

        # if translate_position is set to True, the position
        # supplied will be added to the phsp file position
        if source.translate_position:
            batch[source.position_key_x] += float(source.position.translation[0])
            batch[source.position_key_y] += float(source.position.translation[1])
            batch[source.position_key_z] += float(source.position.translation[2])

        # direction is a rotation of the stored direction
        # if rotate_direction is set to True, the direction
        # in the root file will be rotated based on the supplied rotation matrix
        if source.rotate_direction:
            # create point vectors
            self.points = np.column_stack(
                (
                    batch[source.direction_key_x],
                    batch[source.direction_key_y],
                    batch[source.direction_key_z],
                )
            )
            # create rotation matrix
            r = Rotation.from_matrix(source.position.rotation)
            if source.verbose:
                print("Rotation matrix: ", r.as_matrix())
            # rotate vector with rotation matrix
            points = r.apply(self.points)
            # source.fDirectionX, source.fDirectionY, source.fDirectionZ = points.T
            batch[source.direction_key_x] = points[:, 0].astype(np.float32)
            batch[source.direction_key_y] = points[:, 1].astype(np.float32)
            batch[source.direction_key_z] = points[:, 2].astype(np.float32)

        # set weight
        if source.weight_key != "" and source.weight_key is not None:
            if source.weight_key not in batch:
                fatal(
                    f"PhaseSpaceSource: no Weight key ({source.weight_key}) in the phsp file."
                )
        else:
            self.w = np.ones(current_batch_size, dtype=np.float32)
            batch[source.weight_key] = self.w.astype(np.float32)

        # send to cpp
        # set position
        source.SetPositionXBatch(batch[source.position_key_x])
        source.SetPositionYBatch(batch[source.position_key_y])
        source.SetPositionZBatch(batch[source.position_key_z])

        # set direction
        source.SetDirectionXBatch(batch[source.direction_key_x])
        source.SetDirectionYBatch(batch[source.direction_key_y])
        source.SetDirectionZBatch(batch[source.direction_key_z])

        # set energy
        source.SetEnergyBatch(batch[source.energy_key])

        # set PDGCode
        if source.PDGCode_key in batch:
            source.SetPDGCodeBatch(batch[source.PDGCode_key])
        # set weight
        source.SetWeightBatch(batch[source.weight_key])

        if source.verbose:
            print("PhaseSpaceSourceGenerator: batch generated: ")
            print("particle name: ", source.particle)
            if source.PDGCode_key in batch:
                print("source.fPDGCode: ", batch[source.PDGCode_key])
            print("source.fEnergy: ", batch[source.energy_key])
            print("source.fWeight: ", batch[source.weight_key])
            print("source.fPositionX: ", batch[source.position_key_x])
            print("source.fPositionY: ", batch[source.position_key_y])
            print("source.fPositionZ: ", batch[source.position_key_z])
            print("source.fDirectionX: ", batch[source.direction_key_x])
            print("source.fDirectionY: ", batch[source.direction_key_y])
            print("source.fDirectionZ: ", batch[source.direction_key_z])
            print("source.fEnergy dtype: ", batch[source.energy_key].dtype)

        return current_batch_size


class PhaseSpaceSource(SourceBase, g4.GatePhaseSpaceSource):
    """
    Source of particles from a (root) phase space.
    Read position + direction + energy + weight from the root and use them as event.

    If "global flag" is True, the position/direction are global, ie in the world coordinate system.
    If it is False, it uses the coordinate system of the volume it is attached to.

    The Time in the phsp is not implemented (yet)
    """

    # hints for IDE
    phsp_file: str
    entry_start: int
    particle: str
    global_flag: bool
    isotropic_momentum: bool
    translate_position: bool
    rotate_direction: bool
    batch_size: int
    position_key: str
    position_key_x: str
    position_key_y: str
    position_key_z: str
    direction_key: str
    direction_key_x: str
    direction_key_y: str
    direction_key_z: str
    energy_key: str
    weight_key: str
    PDGCode_key: str
    generate_until_next_primary: bool
    primary_lower_energy_threshold: float
    primary_PDGCode: int
    verbose: bool

    user_info_defaults = {
        "phsp_file": (
            None,
            {"doc": "Filename of the phase-space file (root). This is required"},
        ),
        "entry_start": (
            None,
            {
                "doc": "Starting particle in the phase-space (for MT, provide a list of entries, one for each thread)"
            },
        ),
        "particle": ("", {"doc": "FIXME"}),
        "global_flag": (
            False,
            {
                "doc": "If true, the positions of the generated particles in the phase-space "
                "are in the world coordinate system. If false, they are relative to the volume"
                "this source is attached to",
            },
        ),
        "isotropic_direction": (
            False,
            {
                "doc": "If true, It enables to generate a particle with a position energy and weight "
                "according to the provided phase but with an isotropic momentum.",
            },
        ),
        "translate_position": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "rotate_direction": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "batch_size": (
            10000,
            {
                "doc": "Batch size to read the phsp",
            },
        ),
        "position_key": (
            "PrePositionLocal",
            {
                "doc": "Key in the phsp that contain the position of the particle. Automatically set the"
                " position_key_x, position_key_y, position_key_z keys by adding _X _Y _Z",
                # "setter_hook": _setter_hook_phsp_source_3d_keys,
            },
        ),
        "position_key_x": (
            None,
            {
                "doc": "Key in the phsp that contain the position X of the particle",
            },
        ),
        "position_key_y": (
            None,
            {
                "doc": "Key in the phsp that contain the position Y of the particle",
            },
        ),
        "position_key_z": (
            None,
            {
                "doc": "Key in the phsp that contain the position Z of the particle",
            },
        ),
        "position": (
            Box(
                {"translation": [0, 0, 0], "rotation": Rotation.identity().as_matrix()}
            ),
            {
                "doc": "Default position+rotation if it is not read in the phsp",
            },
        ),
        "direction_key": (
            "PreDirectionLocal",
            {
                "doc": "Key in the phsp that contain the direction of the particle. Automatically set the"
                " direction_key_x, direction_key_y, direction_key_z keys by adding _X _Y _Z",
                # "setter_hook": _setter_hook_phsp_source_3d_keys,
            },
        ),
        "direction_key_x": (
            None,
            {
                "doc": "Key in the phsp that contain the direction X of the particle",
            },
        ),
        "direction_key_y": (
            None,
            {
                "doc": "Key in the phsp that contain the direction Y of the particle",
            },
        ),
        "direction_key_z": (
            None,
            {
                "doc": "Key in the phsp that contain the direction Z of the particle",
            },
        ),
        "energy_key": (
            "KineticEnergy",
            {
                "doc": "Key in the phsp that contain the energy of the particle",
            },
        ),
        "weight_key": (
            "Weight",
            {
                "doc": "Key in the phsp that contain the weight of the particle",
            },
        ),
        "PDGCode_key": (
            "PDGCode",
            {
                "doc": "Key in the phsp that contains the PDGCode of the particle (particle type) "
                "see https://pdg.lbl.gov/2007/reviews/montecarlorpp.pdf",
            },
        ),
        "generate_until_next_primary": (
            False,
            {
                "doc": "FIXME ",
            },
        ),
        "primary_lower_energy_threshold": (
            0,
            {
                "doc": "FIXME ",
            },
        ),
        "primary_PDGCode": (
            0,
            {
                "doc": "FIXME ",
            },
        ),
        "verbose": (
            False,
            {
                "doc": "FIXME ",
            },
        ),
        "verbose_batch": (
            False,
            {
                "doc": "FIXME ",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.__initcpp__()
        # there will be one particle generator per thread
        self.particle_generators = {}
        # number of entries in the phsp root file
        self.num_entries = None

    def __initcpp__(self):
        g4.GatePhaseSpaceSource.__init__(self)

    def __getstate__(self):
        # the particle generator cannot (?) being pickled
        # we convert in a dict with the cycle count values
        all_pg = self.particle_generators
        self.particle_generators = {}
        for k, pg in all_pg.items():
            self.particle_generators[k] = Box({"cycle_count": pg.cycle_count})
        state_dict = super().__getstate__()
        return state_dict

    def initialize(self, run_timing_intervals):
        # create a generator for each thread
        tid = g4.G4GetThreadId()
        self.particle_generators[tid] = PhaseSpaceSourceGenerator(tid)

        # initialize source
        SourceBase.initialize(self, run_timing_intervals)

        # check user info
        if self.position_key_x is None:
            self.position_key_x = f"{self.position_key}_X"
        if self.position_key_y is None:
            self.position_key_y = f"{self.position_key}_Y"
        if self.position_key_z is None:
            self.position_key_z = f"{self.position_key}_Z"

        if self.direction_key_x is None:
            self.direction_key_x = f"{self.direction_key}_X"
        if self.direction_key_y is None:
            self.direction_key_y = f"{self.direction_key}_Y"
        if self.direction_key_z is None:
            self.direction_key_z = f"{self.direction_key}_Z"

        # check if the source should generate particles until the second one
        # which is identified as primary by name, PDGCode and above a threshold
        if self.generate_until_next_primary:
            if self.primary_PDGCode == 0:
                fatal(
                    "PhaseSpaceSource: generate_until_next_primary is True but no primary particle is defined"
                )
            if self.primary_lower_energy_threshold <= 0:
                fatal(
                    "PhaseSpaceSource: generate_until_next_primary is True but no "
                    "primary_lower_energy_threshold is defined"
                )

        # if not set, initialize the entry_start to 0 or to a list for multithreading
        if self.entry_start is None:
            if not g4.IsMultithreadedApplication():
                self.entry_start = 0
            else:
                # create an entry_start array with the correct number of start entries
                # all entries are spaced by the number of particles/thread
                # FIXME: check this line. I corrected it because it seemed like a typo (NK)
                n_threads = self.simulation.number_of_threads
                # n_threads = self.simulation.phsp_source.number_of_threads
                step = np.ceil(self.n / n_threads) + 1  # Specify the increment value
                self.entry_start = [i * step for i in range(n_threads)]

        # initialize the generator (read the phsp file)
        self.particle_generators[tid].initialize(self)

        # keep a copy of the number of entries
        self.num_entries = self.particle_generators[tid].num_entries

        # set the function pointer to the cpp side
        self.SetGeneratorFunction(self.particle_generators[tid].generate)

    @property
    def cycle_count(self):
        if not g4.IsMultithreadedApplication():
            tid = g4.G4GetThreadId()
            return self.particle_generators[tid].cycle_count
        else:
            s = " ".join(
                str(self.particle_generators[tid].cycle_count)
                for tid in self.particle_generators.keys()
            )
            return s


process_cls(PhaseSpaceSource)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
from test085_free_flight_helpers import *
from opengate.contrib.spect.spect_helpers import *

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test085_spect"
    )

    # create the simulation
    sim = gate.Simulation()
    sim.visu = False
    sim.number_of_threads = 4
    ac = 6e3
    source, actors = create_simulation_test085(
        sim,
        paths,
        simu_name="fd",
        ac=ac,
        use_spect_head=True,
        use_spect_arf=False,
        use_phsp=False,
    )

    # FF with Forced Direction (one single head !)
    source.direction.acceptance_angle.intersection_flag = False
    source.direction.acceptance_angle.normal_flag = False
    source.direction.acceptance_angle.forced_direction_flag = True
    source.direction.acceptance_angle.volumes = ["spect_1"]
    source.direction.acceptance_angle.normal_vector = [0, 0, -1]
    source.direction.acceptance_angle.normal_tolerance = 20 * gate.g4_units.deg

    # free flight actor
    ff = sim.add_actor("GammaFreeFlightActor", "ff")
    ff.attached_to = "world"
    ff.unbiased_volumes = ["spect_1_crystal"]

    # go
    sim.run()
    stats = sim.get_actor("stats")
    print(stats)

    # uncertainty
    uncer, _, _ = history_rel_uncertainty_from_files(
        paths.output / "projection_1_fd_counts.mhd",
        paths.output / "projection_1_fd_squared_counts.mhd",
        ac,
        paths.output / "projection_1_fd_uncertainty.mhd",
    )

    # compare to noFF
    is_ok = True
    is_ok = (
        utility.assert_images(
            paths.output_ref / "projection_1_ff_counts.mhd",
            paths.output / "projection_1_fd_counts.mhd",
            stats,
            tolerance=80,
            ignore_value_data1=0,
            sum_tolerance=15,
            sad_profile_tolerance=33,
            scaleImageValuesFactor=2e5 / ac,
            axis="x",
            fig_name=paths.output / "projection_fd_check_1.png",
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.tests.utility as utility

if __name__ == "__main__":
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.verbose_level = gate.logger.DEBUG
    sim.running_verbose_level = 0  # gate.EVENT
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = 1
    sim.random_seed = 13241234

    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.second

    # add a simple volume
    waterbox = sim.add_volume("Box", "Waterbox")
    waterbox.size = [20 * cm, 20 * cm, 20 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 15 * cm]
    waterbox.material = "G4_WATER"

    # default source for tests
    source1 = sim.add_source("GenericSource", "source1")
    source1.particle = "proton"
    source1.energy.mono = 150 * MeV
    source1.position.radius = 10 * mm
    source1.direction.type = "momentum"
    source1.direction.momentum = [0, 0, 1]
    source1.n = 2000

    source2 = sim.add_source("GenericSource", "source2")
    source2.particle = "proton"
    source2.energy.mono = 120 * MeV
    source2.position.radius = 5 * mm
    source2.activity = 2000 * Bq  # 25 + 50 + 100
    source2.direction.type = "momentum"
    source2.direction.momentum = [0, 0, 1]
    source2.start_time = 0.25 * sec

    source3 = sim.add_source("GenericSource", "source3")
    source3.particle = "proton"
    source3.energy.mono = 150 * MeV
    source3.position.radius = 10 * mm
    source3.n = 2400
    source3.start_time = 0.50 * sec
    source3.direction.type = "momentum"
    source3.direction.momentum = [0, 0, 1]

    # Expected total of events
    # 100 + 175 + 120 = 395

    # debug: uncomment to remove one source
    # sim.source_manager.sources.pop('source1')
    # sim.source_manager.sources.pop('source2')
    # sim.source_manager.sources.pop('source3')

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # run timing test #1
    sim.run_timing_intervals = [
        [0, 0.5 * sec],
        [0.5 * sec, 1.0 * sec],
        # Watch out : there is (on purpose) a 'hole' in the timeline
        [1.5 * sec, 2.5 * sec],
    ]

    # print sources
    print(sim.source_manager.dump_sources())

    # start simulation
    sim.run()

    # print stats (useful comment isn't it ?)
    print(stats)

    stats_ref = gate.actors.miscactors.SimulationStatisticsActor(name="stat_ref")
    c = stats_ref.counts
    c.runs = 3
    c.events = 7800
    c.tracks = 37584  # 56394
    c.steps = 266582  # 217234
    # stats_ref.pps = 4059.6 3 3112.2
    c.duration = 1 / 4059.6 * 7800 * sec
    print("-" * 80)
    is_ok = utility.assert_stats(stats, stats_ref, 0.185)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
import numpy as np
import opengate as gate
from matplotlib import pyplot as plt

from opengate import g4_units
from opengate.tests.utility import get_image_1d_profile, print_test


def add_waterbox(sim):
    # units
    cm = gate.g4_units.cm

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [30 * cm, 30 * cm, 20 * cm]
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # low density box
    low_density_box = sim.add_volume("Box", "low_density_box")
    low_density_box.mother = waterbox.name
    low_density_box.size = [30 * cm, 30 * cm, 3 * cm]
    low_density_box.translation = [0 * cm, 0 * cm, -6 * cm]
    low_density_box.material = "G4_lPROPANE"  # density is around 0.43
    low_density_box.color = [0, 1, 1, 1]

    # high density box
    high_density_box = sim.add_volume("Box", "high_density_box")
    high_density_box.mother = waterbox.name
    high_density_box.size = [30 * cm, 30 * cm, 1 * cm]
    high_density_box.translation = [0 * cm, 0 * cm, 1 * cm]
    high_density_box.material = "G4_Pyrex_Glass"  # density is around 2.23
    high_density_box.color = [1, 0, 0, 1]

    return waterbox


def add_simple_waterbox(sim):
    # units
    cm = gate.g4_units.cm

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [30 * cm, 30 * cm, 30 * cm]
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]
    return waterbox


def add_source(
    sim,
    n=1e5,
    energy=350 * g4_units.keV,
    sigma=0.3 * g4_units.MeV,
    radius=3 * g4_units.cm,
):
    cm = gate.g4_units.cm
    source = sim.add_source("GenericSource", "source")
    source.energy.mono = energy
    source.energy.type = "gauss"
    source.energy.sigma_gauss = sigma
    source.particle = "gamma"
    source.position.type = "sphere"
    source.position.radius = radius
    source.position.translation = [0, 0, -55 * cm]
    source.direction.type = "focused"
    source.direction.focus_point = [0, 0, -20 * cm]
    source.n = n / sim.number_of_threads
    if sim.visu:
        source.n = 10
    return source


def add_iso_source(
    sim,
    n=1e5,
    energy=2 * g4_units.MeV,
    sigma=0.6 * g4_units.MeV,
):
    cm = gate.g4_units.cm
    source = sim.add_source("GenericSource", "source")
    source.energy.mono = energy
    source.energy.type = "gauss"
    source.energy.sigma_gauss = sigma
    source.particle = "gamma"
    source.position.type = "sphere"
    source.position.radius = 10 ** (-7) * cm
    source.position.translation = [0, 0, 0]
    source.direction.type = "iso"
    source.n = n / sim.number_of_threads
    if sim.visu:
        source.n = 100
    return source


def voxelize_waterbox(sim, output_folder):
    mm = gate.g4_units.mm
    a = sim.output_dir
    sim.output_dir = output_folder
    sim.voxelize_geometry(
        filename="waterbox_with_inserts_8mm", spacing=(8 * mm, 8 * mm, 8 * mm)
    )
    sim.voxelize_geometry(
        filename="waterbox_with_inserts_5mm", spacing=(5 * mm, 5 * mm, 5 * mm)
    )
    sim.voxelize_geometry(
        filename="waterbox_with_inserts_3mm", spacing=(3 * mm, 3 * mm, 3 * mm)
    )
    sim.voxelize_geometry(
        filename="waterbox_with_inserts_1mm", spacing=(1 * mm, 1 * mm, 1 * mm)
    )
    sim.output_dir = a


def plot_pdd(dose_actor, tle_dose_actor, offset=(0, 0)):
    # plot pdd
    fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(13, 5))

    a = ax[0]
    pdd_x, pdd_y = get_image_1d_profile(
        dose_actor.edep.get_output_path(), "z", offset=offset
    )
    a.plot(pdd_x, pdd_y, label="edep analog")
    pdd_x, pdd_y = get_image_1d_profile(
        tle_dose_actor.edep.get_output_path(), "z", offset=offset
    )
    a.plot(pdd_x, pdd_y, label="edep TLE")
    a.set_xlabel("distance [mm]")
    a.set_ylabel("edep [MeV]")
    a.legend()

    a = a.twinx()
    pdd_x, pdd_y = get_image_1d_profile(
        dose_actor.dose_uncertainty.get_output_path(), "z", offset=offset
    )
    a.plot(
        pdd_x,
        pdd_y,
        label="uncert analog",
        linestyle="--",
        linewidth=0.5,
        color="lightseagreen",
    )
    pdd_x, pdd_y = get_image_1d_profile(
        tle_dose_actor.dose_uncertainty.get_output_path(), "z", offset=offset
    )
    a.plot(
        pdd_x,
        pdd_y,
        label="uncert TLE",
        linestyle="--",
        linewidth=0.5,
        color="darkorange",
    )
    a.legend()

    a = ax[1]
    pdd_x, pdd_y = get_image_1d_profile(
        dose_actor.dose.get_output_path(), "z", offset=offset
    )
    a.plot(pdd_x, pdd_y, label="dose analog")
    pdd_x, pdd_y = get_image_1d_profile(
        tle_dose_actor.dose.get_output_path(), "z", offset=offset
    )
    a.plot(pdd_x, pdd_y, label="dose TLE")
    a.set_xlabel("distance [mm]")
    a.set_ylabel("dose [Gy]")
    a.legend()

    return ax, plt


def mean_rel_diff(arr1, arr2, i1, i2, spacing, ax, tol):
    v1 = arr1[i1:i2].mean()
    v2 = arr2[i1:i2].mean()
    r = np.fabs(v1 - v2) / v1
    b = r < tol
    print_test(
        b,
        f"Relative difference {i1} {i2} {v1:.3f} vs {v2:.3f} "
        f"= {r * 100:.2f}% (tol={tol*100:.0f}) ==> {b}",
    )

    # plot
    ax.plot((i1 * spacing, i2 * spacing), (v1, v1), color="darkgray")
    ax.plot((i1 * spacing, i2 * spacing), (v2, v2), color="black")

    return b


def compare_pdd(f1, f2, spacing, ax, tol, offset=0):
    img1 = itk.imread(f1)
    img2 = itk.imread(f2)
    img1_arr = itk.GetArrayFromImage(img1)
    img2_arr = itk.GetArrayFromImage(img2)
    s = img1_arr.shape
    img1_arr = img1_arr[:, int(s[1] / 2), int(s[2] / 2) + offset]
    img2_arr = img2_arr[:, int(s[1] / 2), int(s[2] / 2) + offset]

    # param of the box
    cm = gate.g4_units.cm
    wbs = 20 * cm  # waterbox size
    ldp = 6 * cm  # low density size
    lds = 3 * cm  # low density position
    hdp = 1 * cm  # high density size
    hds = 1 * cm  # high density position

    # entrance section (water)
    w = int(((-ldp - lds / 2) - (-wbs / 2)) / spacing)
    b = mean_rel_diff(img1_arr, img2_arr, 0, w, spacing, ax, tol)

    # low density section
    l = int(lds / spacing)
    b = mean_rel_diff(img1_arr, img2_arr, w, w + l, spacing, ax, tol) and b

    # water section
    w2 = int(((hdp - hds / 2) - (-ldp + lds / 2)) / spacing)
    a = w + l + w2
    b = mean_rel_diff(img1_arr, img2_arr, w + l, a, spacing, ax, tol) and b

    # high density section
    h = int(hds / spacing)
    b = mean_rel_diff(img1_arr, img2_arr, a, a + h, spacing, ax, tol * 3) and b

    return b



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import opengate.contrib.spect.ge_discovery_nm670 as nm670
import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test051")

    """
    Check the options in DigitizerAdderActor
    - TimeDifference: to store the max time difference within all hits in one single
    - NumberOfHits: store the number of hits in one single
    """

    # create the simulation
    sim = gate.Simulation()

    # units
    m = gate.g4_units.m
    nm = gate.g4_units.nm
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.s
    min = gate.g4_units.min

    # verbose
    sim.visu = False
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.number_of_threads = 1
    sim.random_seed = 123457
    sim.output_dir = paths.output

    # world size
    world = sim.world
    world.size = [2.5 * m, 2.5 * m, 2.5 * m]
    world.material = "G4_AIR"

    # spect head without collimator
    spect, colli, crystal = nm670.add_spect_head(
        sim, "spect", collimator_type=False, debug=False
    )
    nm670.rotate_gantry(spect, 15 * cm, 0)
    # spect.translation = [0, 0, -15 * cm]

    # spect digitizer
    hc = sim.add_actor("DigitizerHitsCollectionActor", f"Hits")
    hc.attached_to = crystal.name
    hc.output_filename = ""  # No output
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]
    sc = sim.add_actor("DigitizerAdderActor", f"Singles")
    sc.attached_to = hc.attached_to
    sc.input_digi_collection = hc.name
    sc.time_difference = True
    sc.number_of_hits = True
    sc.output_filename = "test051_singles.root"

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = True
    sim.physics_manager.global_production_cuts.all = 100 * mm
    sim.physics_manager.set_production_cut("spect", "all", 0.1 * mm)

    # source of Ac225 ion
    s1 = sim.add_source("GenericSource", "source")
    s1.particle = "ion 89 225"  # Ac225
    s1.position.type = "sphere"
    s1.position.radius = 1 * nm
    s1.position.translation = [0, 0, 0]
    s1.direction.type = "iso"
    s1.activity = 1e4 * Bq / sim.number_of_threads

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True
    stats.output_filename = "test051_stats.txt"

    # go
    sim.run(start_new_process=True)

    # check stats
    print()
    gate.exception.warning("Check stats")
    stats_ref = utility.read_stat_file(paths.output_ref / "test051_stats.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.05)

    # check root
    print()
    gate.exception.warning("Check root time difference")
    root1, n1 = utility.open_root_as_np(
        paths.output_ref / "test051_singles.root", "Singles"
    )
    root2, n2 = utility.open_root_as_np(sc.get_output_path(), "Singles")

    # time difference
    td1 = root1["TimeDifference"]
    td2 = root2["TimeDifference"]
    td1 = td1[td1 > 1 * sec] / min
    td2 = td2[td2 > 1 * sec] / min
    is_ok = (
        utility.check_diff_abs(
            len(td1),
            len(td2),
            20,
            f"Number of time diff larger than 1 sec (~{len(td1)/n1*100:.2f}%):",
        )
        and is_ok
    )
    is_ok = (
        utility.check_diff_abs(
            np.mean(td1), np.mean(td2), 35, f"Time diff mean in minutes:"
        )
        and is_ok
    )

    print()
    gate.exception.warning("Check root nb of hits")
    nh1 = root1["NumberOfHits"]
    nh2 = root2["NumberOfHits"]
    is_ok = (
        utility.check_diff(np.mean(nh1), np.mean(nh2), 6, f"Number of hits in mean:")
        and is_ok
    )

    # plot
    f, ax = plt.subplots(1, 2, figsize=(25, 10))
    utility.plot_hist(ax[0], td1, f"Time diff ref (minutes)")
    utility.plot_hist(ax[0], td2, f"Time diff (minutes)")
    utility.plot_hist(ax[1], nh1, f"Nb of hits ref")
    utility.plot_hist(ax[1], nh2, f"Nb of hits")

    fn = paths.output / "test051_singles.png"
    plt.savefig(fn)
    print(f"Plot in {fn}")

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test060_phsp_source_helpers as t
import opengate as gate
from opengate.tests import utility


paths = utility.get_default_test_paths(
    __file__, "test060_PhsSource_ParticleName_direct", output_folder="test060"
)

# units
m = gate.g4_units.m
mm = gate.g4_units.mm
cm = gate.g4_units.cm
nm = gate.g4_units.nm
Bq = gate.g4_units.Bq
MeV = gate.g4_units.MeV
deg = gate.g4_units.deg


def main():
    print("create reference PhS file")
    t.create_test_phs(
        particle="proton",
        phs_name=paths.output / "test_proton_offset",
        number_of_particles=1,
        translation=[10 * cm, 5 * cm, 0 * mm],
    )
    print("testing rotation")
    t.test_source_rotation(
        source_file_name=paths.output / "test_proton_offset.root",
        phs_file_name_out=paths.output / "test_source_rotation.root",
    )

    dirX_ok = t.check_value_from_root_file(
        file_name_root=paths.output / "test_source_rotation.root",
        key="PreDirection_X",
        ref_value=0,
    )
    print("PhS source rotation dirX_ok is correct: ", dirX_ok)

    dirY_ok = t.check_value_from_root_file(
        file_name_root=paths.output / "test_source_rotation.root",
        key="PreDirection_Y",
        ref_value=-0.5,
    )
    print("PhS source rotation dirY_ok is correct: ", dirY_ok)

    dirZ_ok = t.check_value_from_root_file(
        file_name_root=paths.output / "test_source_rotation.root",
        key="PreDirection_Z",
        ref_value=(1.0 - 0.5**2) ** 0.5,
    )
    print("PhS source rotation dirZ_ok is correct: ", dirZ_ok)

    # this is the end, my friend
    utility.test_ok(dirX_ok and dirY_ok and dirZ_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np
import opengate as gate
from opengate.geometry.materials import MaterialDatabase
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "gate_test009_voxels")

    path_to_4d_ct = paths.data / "test070" / "4d_ct"

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.output_dir = paths.output / "test070"

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm
    gcm3 = gate.g4_units.g_cm3
    sec = gate.g4_units.second

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # image
    patient = sim.add_volume("Image", "patient")
    patient.image = path_to_4d_ct / "0.0.mhd"
    patient.load_input_image()
    patient.material = "G4_AIR"  # material used by default
    f1 = str(paths.gate_data / "Schneider2000MaterialsTable.txt")
    f2 = str(paths.gate_data / "Schneider2000DensitiesTable.txt")
    tol = 0.05 * gcm3
    (
        patient.voxel_materials,
        materials,
    ) = gate.geometry.materials.HounsfieldUnit_to_material(sim, tol, f1, f2)
    print(f"tol = {tol} g/cm3")
    print(f"mat : {len(patient.voxel_materials)} materials")
    # cuts
    patient.set_production_cut(particle_name="electron", value=3 * mm)

    # create time intervals to sample a cosine dynamics
    times_cos_first_half = np.arccos(np.linspace(1, -1, 6))[:-1]
    times_cos_second_half = np.pi * 2 - np.arccos(np.linspace(-1, 1, 6))
    times_cos = np.concatenate((times_cos_first_half, times_cos_second_half))
    # normalize tp [0, 1) interval
    times_cos /= 2 * np.pi
    interval_lengths = np.diff(times_cos)

    print(f"times_cos: {times_cos}")
    print(f"interval_lengths: {interval_lengths}")

    # set the run timing intervals
    sim.run_timing_intervals = [
        (t * sec, (t + l) * sec) for t, l in zip(times_cos[:-1], interval_lengths)
    ]

    # set the list of images as dynamic parametrisation
    paths_4d_ct = [path_to_4d_ct / f"{10*i}.0.mhd" for i in range(10)]
    patient.add_dynamic_parametrisation(image=paths_4d_ct)

    # check time samples:
    # y = np.concatenate((np.linspace(1, -1, 6)[:-1], np.linspace(-1, 1, 6)))
    # plt.figure()
    # plt.scatter(times_cos, y)

    # default source for tests
    source = sim.add_source("GenericSource", "proton_source")
    source.energy.mono = 100 * MeV
    source.particle = "proton"
    source.position.type = "sphere"
    source.position.radius = 10 * mm
    source.position.translation = [40 * cm, 0, 0 * cm]
    # source.n = 10000
    source.activity = 10000 * Bq
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test070.mhd"
    dose.attached_to = "patient"
    dose.size = patient.size_pix
    dose.spacing = patient.spacing
    dose.output_coordinate_system = "attached_to_image"
    # dose.translation = [2 * mm, 3 * mm, -2 * mm]
    dose.hit_type = "random"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = False

    # print info
    sim.volume_manager.print_volumes()

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run()

    # print results at the end
    gate.exception.warning(f"Check stats")
    print(stats)
    print(dose)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import uproot
import pathlib
import numpy as np
import itk
import opengate as gate
from opengate.tests import utility


def assert_uncertainty(
    img_E,
    img_err_E,
    nb_part,
    mean_E,
    std_E,
    tab_E,
    tol_th=0.075,
    tol_phsp=0.005,
    is_ok=True,
):
    val_E_img = img_E[0, 0, 0]
    val_err_E_img = round(val_E_img * img_err_E[0, 0, 0], 3)
    val_E_img = round(val_E_img, 2)
    phsp_sum_squared_E = np.sum(tab_E**2)
    phsp_sum_E = np.sum(tab_E)
    phsp_unc = phsp_sum_squared_E / nb_part - (phsp_sum_E / nb_part) ** 2
    phsp_unc = (1 / (nb_part - 1)) * phsp_unc
    phsp_unc = np.sqrt(phsp_unc) * nb_part

    nb_part = int(nb_part)
    print(f"Energy deposited in the voxel for {nb_part} particles : {val_E_img} MeV")
    print(
        f"Energy recorded in the phase space for {nb_part} particles : {round(phsp_sum_E, 2)} MeV"
    )
    print(f"Theoretical deposited energy : {mean_E * nb_part} MeV")
    print(
        f"Standard error on the deposited energy in the voxel for {nb_part} particles : {val_err_E_img} MeV"
    )
    print(
        f"Standard error on the phase space energy deposition {nb_part} particles : {round(phsp_unc, 3)} MeV"
    )

    tstd_e = std_E * np.sqrt(nb_part)
    print(f"Theoretical standard error on deposited energy : {round(tstd_e, 3)} MeV")

    print()
    print(f"Tolerance on the theory comparison: {100 * tol_th}  %")
    print(f"Tolerance on the phase space comparison: {100 * tol_phsp} %")

    var_err_E_th = abs((val_err_E_img - tstd_e)) / tstd_e
    var_err_E_phsp = abs((phsp_unc - val_err_E_img)) / val_err_E_img

    print(
        f"Standard error variation with respect to theory:  {100 * round(var_err_E_th, 3)} %"
    )
    print(
        f"Standard error variation with respect to phase space recording: {100 * round(var_err_E_phsp, 3)} %"
    )

    print(var_err_E_th, var_err_E_phsp)

    if var_err_E_th > tol_th or var_err_E_phsp > tol_phsp:
        is_ok = False

    return is_ok


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test058")
    current_path = pathlib.Path(__file__).parent.resolve()
    output_path = paths.output

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    # sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.EVENT
    sim.number_of_threads = 5
    sim.random_seed = 12365445
    sim.output_dir = output_path

    # units
    m = gate.g4_units.m
    km = gate.g4_units.km
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    gcm3 = gate.g4_units.g_cm3

    #  adapt world size
    world = sim.world
    world.size = [200 * m, 200 * m, 201 * m]

    block_size = [200 * m, 200 * m, 200 * m]

    # Tungsten block
    sim.volume_manager.material_database.add_material_weights(
        "Tungsten",
        ["W"],
        [1],
        19.3 * gcm3,
    )
    t_block = sim.add_volume("Box", "T_block")
    t_block.mother = world.name
    t_block.material = "Tungsten"
    t_block.size = block_size
    t_block.translation = [0 * mm, 0 * mm, -0.5 * m]
    t_block.color = [0, 1, 0, 1]
    t_block.mother = world.name

    # source
    nb_part = 1000 / sim.number_of_threads
    std_dev_E = 10 * keV
    mean_E = 100 * keV
    source = sim.add_source("GenericSource", "photon_source")
    source.particle = "gamma"
    source.n = nb_part
    source.position.type = "box"
    source.position.size = [3 * cm, 3 * cm, 0 * cm]
    source.direction.type = "momentum"
    source.attached_to = world.name
    source.direction.momentum = [0, 0, -1]
    source.energy.type = "gauss"
    source.energy.mono = mean_E
    source.energy.sigma_gauss = std_dev_E

    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    # s.track_types_flag = True

    # add phase space plan

    phsp = sim.add_volume("Box", "phase_space_plane")
    phsp.mother = world.name
    phsp.material = "G4_AIR"
    phsp.size = [200 * m, 200 * m, 1 * nm]
    phsp.translation = [0 * m, 0 * m, 0 * m]
    phsp.color = [1, 0, 0, 1]  # red

    # PhaseSpace Actor
    Phsp_act = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    Phsp_act.attached_to = phsp.name
    Phsp_act.attributes = [
        "KineticEnergy",
        "EventID",
        "ThreadID",
    ]
    Phsp_act.output_filename = "test058_MT.root"
    Phsp_act.debug = False
    Phsp_act.steps_to_store = "exiting first"

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test058_MT.mhd"
    dose.attached_to = t_block.name
    dose.size = [1, 1, 1]
    dose.spacing = block_size
    dose.edep_uncertainty.active = True
    dose.translation = [0 * mm, 0 * mm, -0.5 * m]
    dose.hit_type = "random"

    # Physic list and cuts
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 1 * km
    sim.physics_manager.global_production_cuts.electron = 1 * km
    sim.physics_manager.global_production_cuts.positron = 1 * km

    sim.run()

    # print results
    print(stats)

    # Open images for comparison

    img_E = itk.imread(dose.get_output_path("edep"))
    array_E = itk.GetArrayFromImage(img_E)
    err_img_E = itk.imread(dose.get_output_path("edep_uncertainty"))
    err_array_E = itk.GetArrayFromImage(err_img_E)

    f_phsp = uproot.open(output_path / "test058_MT.root")
    arr_phsp = f_phsp["PhaseSpace"]
    keys_data = arr_phsp.keys(filter_typename="double")
    E = f_phsp["PhaseSpace;1/KineticEnergy"]
    Ephoton = E.array()

    is_ok = assert_uncertainty(
        array_E,
        err_array_E,
        nb_part * sim.number_of_threads,
        mean_E,
        std_dev_E,
        Ephoton,
    )
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
import opengate.contrib.spect.ge_discovery_nm670 as nm670

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, gate_folder="", output_folder="test095"
    )

    # uncomment to regenerate the values
    # nm670.update_geometrical_parameters(store_to_file=True)

    p_ref = nm670.get_geometrical_parameters()
    p = nm670.update_geometrical_parameters()

    for key, v in p.items():
        print(f"{key}: {v}")

    added, removed, modified, same = utility.dict_compare(p, p_ref)
    if added or removed or modified:
        is_ok = False
        print("Error in the geometrical parameters")
        print(f"Added: {added}")
        print(f"Removed: {removed}")
        print(f"Modified: {modified}")
        print(f"Same: {same}")
        for key, v in p_ref.items():
            print(f"{key}: {v}")

    else:
        is_ok = True

    utility.test_ok(is_ok)
